//
//	NOTES:
//		Despite the proposed SDF spec, the SDF images that are floating around don't actually
//		have the header on top of the format.  When properly implemented I expect it'll have
//		to cope with those files in addition to the real versions.  I've left out the SDF
//		saving until the header is put it, to avoid any more headerless files floating about!
//
//		The CFloppyDisk class is a platform specific implementation - easy on some versions
//		of Unix, but messy on MS Windows. See the OS specific Floppy.* files.

#include "stdafx.h"

#include "Util.h"
#include "CDisk.h"
#include "Floppy.h"

////////////////////////////////////////////////////////////////////////////////

/*static*/ CDisk* CDisk::Open (const char* pcszDisk_, bool fReadOnly_/*=false*/)
{
	// First of all try and get a data stream for the disk
	CStream* pStream = CStream::Open(pcszDisk_, fReadOnly_);

	// A disk will only be returned if the stream format is recognised
	CDisk* pDisk = NULL;

	if (pStream)
	{
		// Is it a real floppy disk? (OS-specific implementation)
		if (CFloppyDisk::IsRecognised(pStream))
			pDisk = new CFloppyDisk(pStream);

		// Is it a .SDF image?
		else if (CSDFDisk::IsRecognised(pStream))
			pDisk = new CSDFDisk(pStream);

		// Is it a .SAD image?
		else if (CSADDisk::IsRecognised(pStream))
			pDisk = new CSADDisk(pStream);

		// Is it a .DSK image?
		else if (CDSKDisk::IsRecognised(pStream))
			pDisk = new CDSKDisk(pStream);
	}

	// Return the disk pointer, or NULL if we didn't recognise it
	return pDisk;
}


CDisk::CDisk (CStream* pStream_)
{
	// Save the supplied stream
	m_pStream = pStream_;

	// No geometry information, current track/size or spin position
	m_nSides = m_nTracks = m_nSectors = m_nSector = 0;
	m_nSide = m_nTrack = -1;
	m_nSpinPos = 1;

	// No sector data available
	m_pbData = NULL;

	// Disk not modified, yet
	m_fModified = false;

}

/*virtual*/ CDisk::~CDisk ()
{
	// Delete the stream object and disk data memory we allocated
	if (m_pStream) { delete m_pStream; m_pStream = NULL; }
	if (m_pbData) { delete m_pbData; m_pbData = NULL; }
}


// Sector spin position on the spinning disk, as used by the READ_ADDRESS command
int CDisk::GetSpinPos (bool fAdvance_/*=false*/)
{
	// If required, move to the next spin position for next time
	if (fAdvance_)
		m_nSpinPos = (m_nSpinPos % (m_nSectors + !m_nSectors)) + 1;

	// Return the current/new position
	return m_nSpinPos;
}

// Initialise a sector enumeration, return the number of sectors on the track
int CDisk::FindInit (int nSide_, int nTrack_)
{
	// Remember the current side and track number
	m_nSide = nSide_;
	m_nTrack = nTrack_;

	// Set current sector to zero so the next sector is sector 1
	m_nSector = 0;

	// Return the number of sectors on the current track
	return (m_nSide < m_nSides && m_nTrack < m_nTracks) ? m_nSectors : 0;
}

bool CDisk::FindNext (IDFIELD* pIdField_/*=NULL*/, BYTE* pbStatus_/*=NULL*/)
{
	// Return true if there is another item to be found
	return (m_nSide < m_nSides && m_nTrack < m_nTracks) && ++m_nSector <= m_nSectors;
}

////////////////////////////////////////////////////////////////////////////////

CDSKDisk::CDSKDisk (CStream* pStream_) : CDisk(pStream_)
{
	pStream_->Rewind();

	// The DSK geometry is fixed
	m_nSides = NORMAL_DISK_SIDES;
	m_nTracks = NORMAL_DISK_TRACKS;
	m_nSectors = NORMAL_DISK_SECTORS;
	m_nSectorSize = NORMAL_SECTOR_SIZE;

	// Work out the disk size, allocate some memory for it and read it in
	int nDiskSize = m_nSides * m_nTracks * m_nSectors * m_nSectorSize;
	pStream_->Read(m_pbData = new BYTE[nDiskSize], nDiskSize);
}

/*static*/ bool CDSKDisk::IsRecognised (CStream* pStream_)
{
	char ab[DSK_IMAGE_SIZE];

	// It's only a DSK image if it's exactly the right size, since there's no signature to check for
	return pStream_->Rewind() && pStream_->Read(ab, sizeof ab) == sizeof ab && !pStream_->Read(ab, 1);
}


// Find the next sector in the current track
bool CDSKDisk::FindNext (IDFIELD* pIdField_, BYTE* pbStatus_)
{
	bool fRet = CDisk::FindNext();

	// Make sure there is a 'next' one
	if (fRet)
	{
		// Construct a normal ID field for the sector
		pIdField_->bSide = m_nSide;
		pIdField_->bTrack = m_nTrack;
		pIdField_->bSector = m_nSector;
		pIdField_->bSize = 2;						// 128 << 2 = 512 bytes
		pIdField_->bCRC1 = pIdField_->bCRC2 = 0;

		// Sector OK so reset all errors
		*pbStatus_ = 0;
	}

	// Return true if we're returning a new item
	return fRet;
}

// Read the data for the last sector found
BYTE CDSKDisk::ReadData (BYTE *pbData_, UINT* puSize_)
{
	// Work out the offset for the required data
	long lPos = (m_nSide + 2 * m_nTrack) * (m_nSectors * NORMAL_SECTOR_SIZE) + ((m_nSector-1) * NORMAL_SECTOR_SIZE);

	// Copy the sector data from the image buffer
	memcpy(pbData_, m_pbData + lPos, *puSize_ = NORMAL_SECTOR_SIZE);

	// Data is always perfect on DSK images, so return OK
	return 0;
}

// Read the data for the last sector found
BYTE CDSKDisk::WriteData (BYTE *pbData_, UINT* puSize_)
{
	// Work out the offset for the required data
	long lPos = (m_nSide + 2 * m_nTrack) * (m_nSectors * NORMAL_SECTOR_SIZE) + ((m_nSector-1) * NORMAL_SECTOR_SIZE);

	// Copy the sector data to the image buffer, and set the modified flag
	memcpy(m_pbData + lPos, pbData_, *puSize_ = NORMAL_SECTOR_SIZE);
	m_fModified = true;

	// Data is always perfect on DSK images, so return OK
	return 0;
}

bool CDSKDisk::Save ()
{
	// Write the image out as a single block
	if (m_pStream->Rewind() && m_pStream->Write(m_pbData, DSK_IMAGE_SIZE) == DSK_IMAGE_SIZE)
	{
		m_fModified = false;
		return true;
	}

	TRACE("!!! CDSKDisk::Save() failed to write modified disk contents!\n");
	return false;
}

////////////////////////////////////////////////////////////////////////////////

/*static*/ bool CSADDisk::IsRecognised (CStream* pStream_)
{
	SAD_HEADER sh;

	// Read the header, check for the signature, and make sure the disk geometry is sensible
	return (pStream_->Rewind() && pStream_->Read(&sh, sizeof sh) == sizeof sh &&
			!memcmp(sh.abSignature, SAD_SIGNATURE, sizeof sh.abSignature) &&
			sh.bSides && sh.bSides <= MAX_DISK_SIDES && sh.bTracks && sh.bTracks < MAX_DISK_TRACKS &&
			sh.bSectorSizeDiv64 && sh.bSectorSizeDiv64 <= 16 &&
			(sh.bSectorSizeDiv64 & -sh.bSectorSizeDiv64) == sh.bSectorSizeDiv64);
}

CSADDisk::CSADDisk (CStream* pStream_) : CDisk(pStream_)
{
	SAD_HEADER sh;

	pStream_->Rewind();

	// Read the header and extract the disk geometry
	pStream_->Read(&sh, sizeof sh);
	m_nSides = sh.bSides;
	m_nTracks = sh.bTracks;
	m_nSectors = sh.bSectors;
	m_nSectorSize = sh.bSectorSizeDiv64 << 6;

	// Work out the disk size, allocate some memory for it and read it in
	int nDiskSize = sizeof sh + m_nSides * m_nTracks * m_nSectors * m_nSectorSize;
	memcpy(m_pbData = new BYTE[nDiskSize], &sh, sizeof sh);
	pStream_->Read(m_pbData + sizeof sh, nDiskSize);
}

// Find the next sector in the current track
bool CSADDisk::FindNext (IDFIELD* pIdField_, BYTE* pbStatus_)
{
	bool fRet = CDisk::FindNext();

	// Make sure there is a 'next' one
	if (fRet)
	{
		// Construct a normal ID field for the sector
		pIdField_->bSide = m_nSide;
		pIdField_->bTrack = m_nTrack;
		pIdField_->bSector = m_nSector;
		pIdField_->bSize = 2;						// 128 << 2 = 512 bytes
		pIdField_->bCRC1 = pIdField_->bCRC2 = 0;

		// Sector OK so reset all errors
		*pbStatus_ = 0;
	}

	// Return true if we're returning a new item
	return fRet;
}

// Read the data for the last sector found
BYTE CSADDisk::ReadData (BYTE *pbData_, UINT* puSize_)
{
	// Work out the offset for the required data
	long lPos = sizeof(SAD_HEADER) + (m_nSide * m_nTracks + m_nTrack) * (m_nSectors * NORMAL_SECTOR_SIZE) + ((m_nSector-1) * NORMAL_SECTOR_SIZE) ;

	// Copy the sector data from the image buffer
	memcpy(pbData_, m_pbData + lPos, *puSize_ = NORMAL_SECTOR_SIZE);

	// Data is always perfect on SAD images, so return OK
	return 0;
}

// Read the data for the last sector found
BYTE CSADDisk::WriteData (BYTE *pbData_, UINT* puSize_)
{
	// Work out the offset for the required data
	long lPos = sizeof(SAD_HEADER) + (m_nSide * m_nTracks + m_nTrack) * (m_nSectors * NORMAL_SECTOR_SIZE) + ((m_nSector-1) * NORMAL_SECTOR_SIZE) ;

	// Copy the sector data to the image buffer, and set the modified flag
	memcpy(m_pbData + lPos, pbData_, *puSize_ = NORMAL_SECTOR_SIZE);
	m_fModified = true;

	// Data is always perfect on SAD images, so return OK
	return 0;
}

bool CSADDisk::Save ()
{
	int nDiskSize = sizeof SAD_HEADER + m_nSides * m_nTracks * m_nSectors * m_nSectorSize;

	if (m_pStream->Rewind() && m_pStream->Write(m_pbData, nDiskSize) == nDiskSize)
	{
		m_fModified = false;
		return true;
	}

	TRACE("!!! SADDisk::Save() failed to write modified disk contents!\n");
	return false;
}

////////////////////////////////////////////////////////////////////////////////

/*static*/ bool CSDFDisk::IsRecognised (CStream* pStream_)
{
	bool fValid = false;

	BYTE* pb = new BYTE [OLD_SDF_IMAGE_SIZE];
	if (pb)
	{
		SDF_HEADER sh;

		// Old files have no header and are a fixed size
		if (pStream_->Rewind() && (pStream_->Read(pb, OLD_SDF_IMAGE_SIZE) == OLD_SDF_IMAGE_SIZE) && !pStream_->Read(pb, 1))
			fValid = true;

		// Read the header
		else if (pStream_->Rewind() && pStream_->Read(&sh, sizeof sh) == sizeof sh)
		{
			// Check for the signature, and make sure the geometry is sensible
			if (!memcmp(&sh.abSignature, SDF_SIGNATURE, sizeof sh.abSignature) &&
				sh.bSides && sh.bSides <= MAX_DISK_SIDES && sh.bTracks && sh.bTracks <= MAX_DISK_TRACKS)
				fValid = true;
		}

		delete pb;
	}

	return fValid;
}

CSDFDisk::CSDFDisk (CStream* pStream_) : CDisk(pStream_)
{
	pStream_->Rewind();

	// Read the header (if any)
	SDF_HEADER sh;
	pStream_->Read(&sh, sizeof sh);

	// If we find the disk signature, extract the disk geometry from it
	if (!memcmp(&sh.abSignature, SDF_SIGNATURE, sizeof sh.abSignature))
	{
		m_nSides = sh.bSides;
		m_nTracks = sh.bTracks;
	}

	// Else we have an old format without a header
	else
	{
		// Set up the fixed geometry used by old version
		m_nSides = NORMAL_DISK_SIDES;
		m_nTracks = NORMAL_DISK_TRACKS;

		// Rewind to the start to the data again
		pStream_->Rewind();

		// If the stream isn't read-only, mark the disk as modified so it gets written back in the new format
		if (!pStream_->IsReadOnly())
			SetModified();
	}

	// SDF images don't have a fixed number of sectors per track or sector size
	m_nSectors = 0;
	m_nSectorSize = 0;

	// Work out the disk size, allocate some memory for it and read it in
	int nDiskSize = m_nSides * m_nTracks * SDF_TRACKSIZE;
	pStream_->Read(m_pbData = new BYTE[nDiskSize], nDiskSize);
}


// Initialise the enumeration of all sectors in the track
int CSDFDisk::FindInit (int nSide_, int nTrack_)
{
	if (nSide_ >= m_nSides || nTrack_ >= m_nTracks)
		return m_nSectors = 0;

	// Locate the track in the buffer to determine the number of sectors available
	int nPos = /*sizeof SDF_HEADER*/ + (SDF_TRACKSIZE)*(nSide_*m_nTracks + nTrack_);
	m_pTrack = reinterpret_cast<TRACK_HEADER*>(m_pbData+nPos);
	m_nSectors = m_pTrack->bSectors;

	m_pFind = NULL;

	// Call the base and return the number of sectors in the track
	return CDisk::FindInit(nSide_, nTrack_);
}

// Find the next sector in the current track
bool CSDFDisk::FindNext (IDFIELD* pIdField_, BYTE* pbStatus_)
{
	bool fRet;

	// Make sure there is a 'next' one
	if (fRet = CDisk::FindNext())
	{
		if (!m_pFind)
			m_pFind = &m_pTrack->sHeader1;
		else
		{
			// The data length is zero if there is a problem with the ID header
			int nSize = m_pFind->bIdStatus ? 0 : (MIN_SECTOR_SIZE << m_pFind->sIdField.bSize);

			// Advance the pointer over the header and data to next header
			m_pFind = reinterpret_cast<SECTOR_HEADER*>(reinterpret_cast<BYTE*>(m_pFind) + nSize + sizeof(SECTOR_HEADER));
		}

		// Copy the ID field to the supplied buffer, and store the ID status
		memcpy(pIdField_, &m_pFind->sIdField, sizeof(*pIdField_));
		*pbStatus_ = m_pFind->bIdStatus;
	}

	// Return true if we're returning a new item
	return fRet;
}

// Read the data for the last sector found
BYTE CSDFDisk::ReadData (BYTE *pbData_, UINT* puSize_)
{
	// Copy the sector data
	memcpy(pbData_, reinterpret_cast<BYTE*>(m_pFind+1), *puSize_ = MIN_SECTOR_SIZE << m_pFind->sIdField.bSize);

	// Return the data status to include a possible CRC error
	return m_pFind->bDataStatus;
}

// Read the data for the last sector found
BYTE CSDFDisk::WriteData (BYTE *pbData_, UINT* puSize_)
{
	// Copy the sector data
	memcpy(reinterpret_cast<BYTE*>(m_pFind+1), pbData_, *puSize_ = MIN_SECTOR_SIZE << m_pFind->sIdField.bSize);
	m_fModified = true;

	// Written ok
	return 0;
}

bool CSDFDisk::Save ()
{
	if (IsModified())
	{
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////


CFloppyDisk::CFloppyDisk (CStream* pStream_) : CDisk(pStream_)
{
	// We can assume this as only CFloppyStream will recognise the real device, and we are always the disk used
	m_pFloppy = reinterpret_cast<CFloppyStream*>(pStream_);

	// Maximum supported geometry
	m_nSides = 2;
	m_nTracks = 83;
	m_nSectors = 10;
	m_nSectorSize = 512;

	// Work out the disk size, allocate some memory for it and read it in
	int nDiskSize = m_nSides * m_nTracks * m_nSectors * m_nSectorSize;
	m_pbData = new BYTE[nDiskSize];

	// Clear the track cache as no tracks are present
	memset(m_abTrackCache, 0, sizeof(m_abTrackCache));
}

/*static*/ bool CFloppyDisk::IsRecognised (CStream* pStream_)
{
	return CFloppyStream::IsRecognised(pStream_->GetName());
}


// Find the next sector in the current track
bool CFloppyDisk::FindNext (IDFIELD* pIdField_, BYTE* pbStatus_)
{
	bool fRet = CDisk::FindNext();

	// Make sure there is a 'next' one
	if (fRet)
	{
		// Construct a normal ID field for the sector
		pIdField_->bSide = m_nSide;
		pIdField_->bTrack = m_nTrack;
		pIdField_->bSector = m_nSector;
		pIdField_->bSize = 2;						// 128 << 2 = 512 bytes
		pIdField_->bCRC1 = pIdField_->bCRC2 = 0;

		// Sector OK so reset all errors
		*pbStatus_ = 0;
	}

	// Return true if we're returning a new item
	return fRet;
}

// Read the data for the last sector found
BYTE CFloppyDisk::ReadData (BYTE *pbData_, UINT* puSize_)
{
	return m_pFloppy->Read(m_nSide, m_nTrack, m_nSector, pbData_, puSize_);
}

// Read the data for the last sector found
BYTE CFloppyDisk::WriteData (BYTE *pbData_, UINT* puSize_)
{
	return m_pFloppy->Write(m_nSide, m_nTrack, m_nSector, pbData_, puSize_);
}

bool CFloppyDisk::Save ()
{
	// Eveything's currently written immediately, but we'll need to do more if/when writes are cached
	return true;
}
