#ifndef CDISK_H
#define CDISK_H

#include "VL1772.h"		// for the VL-1772 controller definitions
#include "CStream.h"	// for the data stream abstraction
#include "Floppy.h"

////////////////////////////////////////////////////////////////////////////////

#define MAX_DISK_SIDES			2		// Maximum number of sides for a disk image
#define MAX_DISK_TRACKS			83		// Maxmimum number of tracks per side (docs say 0 to 244?)
#define MIN_SECTOR_SIZE			128		// Minimum sector size in bytes
#define MAX_SECTOR_SIZE			1024	// Maximum sector size in bytes

#define NORMAL_DISK_SIDES		2		// Normally 2 sides per disk
#define NORMAL_DISK_TRACKS		80		// Normally 80 tracks per side
#define NORMAL_DISK_SECTORS		10		// Normally 10 sectors per track
#define NORMAL_SECTOR_SIZE		512		// Normally 512 bytes per sector

// The SDF track size needs to be enough to hold the data for any possible SAM format, along with associated sector header data
#define SDF_TRACKSIZE			(NORMAL_SECTOR_SIZE * 12)

#define	MSDOS_DISK_SECTORS		9		// MS-DOS disks are normally 9 sectors per track


// The various disk format image sizes
#define DSK_IMAGE_SIZE			(NORMAL_DISK_SIDES * NORMAL_DISK_TRACKS * NORMAL_DISK_SECTORS * NORMAL_SECTOR_SIZE)
#define MSDOS_IMAGE_SIZE		(NORMAL_DISK_SIDES * NORMAL_DISK_TRACKS * MSDOS_DISK_SECTORS * NORMAL_SECTOR_SIZE)
#define OLD_SDF_IMAGE_SIZE		(NORMAL_DISK_SIDES * NORMAL_DISK_TRACKS * SDF_TRACKSIZE)

#define SDF_SIGNATURE			"SDF"

////////////////////////////////////////////////////////////////////////////////

// The ID string for Aley Keprt's SAD disk image format  (heh)
#define SAD_SIGNATURE			"Aley's disk backup"

// Format of a SAD image header
typedef struct
{
	BYTE	abSignature[sizeof SAD_SIGNATURE - 1];

	BYTE	bSides;				// Number of sides on the disk
	BYTE	bTracks;			// Number of tracks per side
	BYTE	bSectors;			// Number of sectors per track
	BYTE	bSectorSizeDiv64;	// Sector size divided by 64
}
SAD_HEADER;

////////////////////////////////////////////////////////////////////////////////

// Sector header on each sector in the SDF disk image
typedef struct
{
	BYTE	bIdStatus;		// Status error bits for after READ_ADDRESS command
	BYTE	bDataStatus;	// Status error bits for after READ_XSECTOR commands

	IDFIELD	sIdField;		// ID field detailing sector

//	BYTE	abData[];		// Start of sector data (only if bIdStatus is zero)
}
SECTOR_HEADER;


// Track header on each track in the SDF disk image
typedef struct
{
	BYTE	bSectors;		// Number of sectors on this track

	SECTOR_HEADER sHeader1;	// First of the sector headers for this track
}
TRACK_HEADER;


// SDF file header - not present on very old images!
typedef struct
{
	BYTE	abSignature[3];		// SDF image signature ('S' 'D' 'F')
	BYTE	bRevision;			// Compatability number, incremented if/when the header is changed/enhanced

	BYTE	bTracks;			// Tracks per side (usually 80)
	BYTE	bSides;				// Sides (usually 2)

	BYTE	abReserved[12];		// Reserved for future expansion
}
SDF_HEADER;


class CDisk
{
	// Constructor and virtual destructor
	public:
		CDisk (CStream* pStream_);
		virtual ~CDisk ();

	public:
		static CDisk* Open (const char* pcszDisk_, bool fReadOnly_=false);

	// Public query functions
	public:
		const char* GetName () { return m_pStream->GetName(); }
		int GetSpinPos (bool fAdvance_=false);
		bool IsReadOnly () const { return m_pStream->IsReadOnly(); }
		bool IsModified () const { return m_fModified; }

	// Protected overrides
	public:
		virtual int FindInit (int nSide_, int nTrack_);
		virtual bool FindNext (IDFIELD* pIdField_=NULL, BYTE* pbStatus_=NULL) = 0;
		virtual BYTE ReadData (BYTE* pbData_, UINT* puSize_) = 0;
		virtual BYTE WriteData (BYTE* pbData_, UINT* puSize_) = 0;
		virtual bool Save () = 0;

	protected:
		int		m_nSides, m_nTracks, m_nSectors, m_nSectorSize;
		int		m_nSide, m_nTrack, m_nSector;
		bool	m_fModified;

		int		m_nSpinPos;
		CStream*m_pStream;
		BYTE*	m_pbData;

	protected:
		void SetModified () { m_fModified = true; }
		
};


class CDSKDisk : public CDisk
{
	public:
		CDSKDisk (CStream* pStream_);
		virtual ~CDSKDisk () { if (IsModified()) Save(); }

	public:
		static bool IsRecognised (CStream* pStream_);

	public:
		bool FindNext (IDFIELD* pIdField_, BYTE* pbStatus_);
		BYTE ReadData (BYTE* pbData_, UINT* puSize_);
		BYTE WriteData (BYTE* pbData_, UINT* puSize_);
		bool Save ();
};


class CSADDisk : public CDisk
{
	public:
		CSADDisk (CStream* pStream_);
		virtual ~CSADDisk () { if (IsModified()) Save(); }

	public:
		static bool IsRecognised (CStream* pStream_);

	public:
		bool FindNext (IDFIELD* pIdField_, BYTE* pbStatus_);
		BYTE ReadData (BYTE* pbData_, UINT* puSize_);
		BYTE WriteData (BYTE* pbData_, UINT* puSize_);
		bool Save ();
};


class CSDFDisk : public CDisk
{
	public:
		CSDFDisk (CStream* pStream_);
		virtual ~CSDFDisk () { if (IsModified()) Save(); }

	public:
		static bool IsRecognised (CStream* pStream_);

	public:
		int FindInit (int nSide_, int nTrack_);
		bool FindNext (IDFIELD* pIdField_, BYTE* pbStatus_);
		BYTE ReadData (BYTE* pbData_, UINT* puSize_);
		BYTE WriteData (BYTE* pbData_, UINT* puSize_);
		bool Save ();

	protected:
		TRACK_HEADER* m_pTrack;		// Last track 
		SECTOR_HEADER* m_pFind;		// Last sector found with FindNext()
};


class CFloppyDisk : public CDisk
{
	public:
		CFloppyDisk (CStream* pStream_);

	public:
		static bool IsRecognised (CStream* pStream_);

	public:
		bool FindNext (IDFIELD* pIdField_, BYTE* pbStatus_);
		BYTE ReadData (BYTE* pbData_, UINT* puSize_);
		BYTE WriteData (BYTE* pbData_, UINT* puSize_);
		bool Save ();

	protected:
		BYTE m_abTrackCache[MAX_DISK_SIDES * MAX_DISK_TRACKS];	// will do until probably converted to bit array
		CFloppyStream* m_pFloppy;
};

#endif	// CDISK_H
