//	Notes:
//		This is where all the video generation is done for the various SAM
//		modes and host colour depths.  Template classes are used to avoid
//		having to maintain multiple versions, tho that has meant 24-bit mode
//		isn't supported yet - may be done as a hack on 32-bit.
//
//		The update functions can draw partial lines, so we can correctly draw
//		what is seen when the screen mode is changed part way along lines, etc.
//		I did originally keep the functions to do complete lines, but it didn't
//		seem _that_ much slower to use the partial ones, so I got rid of them.
//
//		The DOUBLEX template parameters allows selection between 1x1 (for
//		256x192 modes) and 2x1 (for 512x192 modes). YUV generates the image
//		in YUV format for cards only supporting YUV overlays.
//
//		Note that the above parameters appear to be tested during run-time
//		but since they're template parameters a complete class is generated
//		from each case used, and will end up as  if (true)  etc. and be
//		optimised out by the compiler, leaving just the code needed.

#ifndef DISPLAY_H
#define DISPLAY_H

#include "CPU.h"
#include "Util.h"
#include "IO.h"
#include "Options.h"
#include "Profile.h"


// Helper macros to map a byte in the attribute area to a colour index
#define ATTR_BG(a)	((((a) >> 3) & 7) | (((a) >> 3) & 8))
#define ATTR_FG(a)	(((a) & 7) | (((a) >> 3) & 8))


extern unsigned int y_dtbl[];

extern bool g_fFlashPhase;
extern bool fDrawFrame;

extern BYTE *apbPageReadPtrs[],  *apbPageWritePtrs[];

extern int g_nLastLineNo, g_nLastBlockNo;	// Where we've drawn up to so far


// Template flags so we don't need lots of extra parameters
#define F_DOUBLEX			0x01		// Set to produce a 2x1 image with mode 3 correct
#define F_YUV				0x02		// Set to generate the image in YUV format, as needed for the overlays on some cards
#define F_PALETTE			0x04		// Set to generate the image in SAM palette colour format instead of as native pixels


// Generic base for all screen classes
class CDisplay
{
	typedef void (CDisplay::* FNLINEUPDATE)(int nLine_, int nFrom_, int nTo_);

	public:
		CDisplay ()
		{
			m_nMode = 1;
			m_pLineUpdate = &CDisplay::Mode1Line;

			m_nLine = 0;
			g_nLastLineNo = g_nLastBlockNo = 0;
		}

	public:
		void Update (int nLine_=HEIGHT_LINES, int nScanPos_=TSTATES_PER_LINE)
		{
			if (!fDrawFrame)
				return;

			ProfileStart(GFX);

			int nBlock = nScanPos_ >> 3;

			// If we're still on the same line as last time we just need to update part of a line
			if (nLine_ == g_nLastLineNo)
			{
				UpdateLine(nLine_, g_nLastBlockNo, nBlock);
				g_nLastBlockNo = nBlock;
			}

			// We've got a block to update
			else
			{
				// Finish the rest of the line we updated
				UpdateLine(g_nLastLineNo, g_nLastBlockNo, WIDTH_BLOCKS);

				// Update any complete lines in the middle
				for (int nLine = g_nLastLineNo+1 ; nLine < nLine_ ; nLine++)
					UpdateLine(nLine, 0, WIDTH_BLOCKS);

				// Update to where we've reached on the current line
				UpdateLine(nLine_, 0, nBlock);

				// If we've just finished the frame, prepare for next time
				if (nLine_ == HEIGHT_LINES && nBlock == WIDTH_BLOCKS)
					g_nLastLineNo = g_nLastBlockNo = 0;

				// otherwise remember where we've drawn up to so we can continue next time
				else
				{
					g_nLastLineNo = nLine_;
					g_nLastBlockNo = nBlock;
				}
			}

			ProfileEnd();
		}

		// Spectrum compatability attribute value
		BYTE GetAttr ()
		{
			if (m_nLine < TOP_BORDER_LINES && m_nLine >= (TOP_BORDER_LINES+SCREEN_LINES))
				return 0xff;
			return 0;

			// More mode checking here?
		}

		BYTE GetHPen (int nLine_)
		{
			// Return 192 for the border areas, or the real line number
			// Note: the right-border area is treated as part of the following line
			return (nLine_ < TOP_BORDER_LINES || nLine_ >= (TOP_BORDER_LINES+SCREEN_LINES)) ? (BYTE)SCREEN_LINES :
					nLine_ - TOP_BORDER_LINES + (g_nLineCycle > (BORDER_PIXELS+SCREEN_PIXELS-VIDEO_DELAY));
		}

		void SetMode (int nMode_, int nLine_=-1, int nScanPos_=-1)
		{
			if (fDrawFrame && nLine_ != -1)
			{
				// If the mode changes from 1/2 to 3/4 or vice versa, we need to show the strange ASIC block generated by
				// using the data from one mode in another mode (well, sorta) - the Laundon effect! :-)
				if (((m_nMode-1) ^ (nMode_-1)) & 0x02)
				{
					ModeChange(nMode_, nLine_-OFFSCREEN_LINES, nScanPos_ >> 3);
					g_nLastBlockNo += (VIDEO_DELAY >> 3);
				}
			}

			static FNLINEUPDATE apfnLineUpdates[] =
				{ &CDisplay::Mode1Line, &CDisplay::Mode2Line, &CDisplay::Mode3Line, &CDisplay::Mode4Line };

			m_pLineUpdate = apfnLineUpdates[(m_nMode = nMode_)-1];

			// VMPR page must be even for modes 3 and 4
			int nPage = nMode_ >= 3 ? (VMPR_SCREEN_PAGE & 0x1e) : VMPR_SCREEN_PAGE;
			m_pbScreenData = apbPageReadPtrs[nPage];
		}

	protected:
		void UpdateLine (int nLine_, int nFrom_, int nTo_)
		{
			m_nLine = nLine_;

			// Does the line require drawing?
			if (nLine_ >= OFFSCREEN_LINES && nLine_ < (TOP_BORDER_LINES+SCREEN_LINES+VISIBLE_LINES))
			{
				// The screen off bit only applies to modes 3 and 4
				if (BORD_SOFF && m_nMode >= 3)
					BlackLine(nLine_-OFFSCREEN_LINES, nFrom_, nTo_);

				// If the line is on the main screen, call the current mode function to draw it
				else if (nLine_ >= TOP_BORDER_LINES && nLine_ < (TOP_BORDER_LINES+SCREEN_LINES))
					(this->*m_pLineUpdate)(nLine_-OFFSCREEN_LINES, nFrom_, nTo_);

				// Draw the top or bottom border if necessary
				else if ((nLine_ >= OFFSCREEN_LINES && nLine_ < TOP_BORDER_LINES) ||
					(nLine_ >= (TOP_BORDER_LINES+SCREEN_LINES) && nLine_ < (TOP_BORDER_LINES+SCREEN_LINES+VISIBLE_LINES)))
					BorderLine(nLine_-OFFSCREEN_LINES, nFrom_, nTo_);
			}
		}

	protected:
		virtual void BorderLine (int nLine_, int nFrom_, int nTo_) = 0;
		virtual void BlackLine (int nLine_, int nFrom_, int nTo_) = 0;

		virtual void Mode1Line (int nLine_, int nFrom_, int nTo_) = 0;
		virtual void Mode2Line (int nLine_, int nFrom_, int nTo_) = 0;
		virtual void Mode3Line (int nLine_, int nFrom_, int nTo_) = 0;
		virtual void Mode4Line (int nLine_, int nFrom_, int nTo_) = 0;

		virtual void ModeChange (int nMode_, int nLine_, int nBlock_) = 0;

	protected:
		FNLINEUPDATE m_pLineUpdate;						// Function used to draw current mode
		BYTE*	m_pbScreenData;
		int		m_nMode;								// Current SAM screen mode
		int		m_nLine;								// Line number being drawn

	public:
		BYTE* m_pbFrameStart;		// TEMPORARY - TAKE OUT!
};


////////////////////////////////////////////////////////////////////////////////
//	C S c r e e n X x 1

// Virtual base class for all screen depths
template <class T, BYTE bFlags_>
class CDisplayXx1 : public CDisplay
{
	public:
		CDisplayXx1 (BYTE* pbFrame_=NULL, LONG lPitch_=0);

	protected:
		void Mode1Line (int nLine_, int nFrom_, int nTo_);
		void Mode2Line (int nLine_, int nFrom_, int nTo_);
		void Mode3Line (int nLine_, int nFrom_, int nTo_);
		void Mode4Line (int nLine_, int nFrom_, int nTo_);
		void ModeChange (int nMode_, int nLine_, int nBlock_);

	protected:
		void LeftBorder (int nFrom_, int nTo_);
		void RightBorder (int nFrom_, int nTo_);
		void BorderLine (int nLine_, int nFrom_, int nTo_);
		void BlackLine (int nLine_, int nFrom_, int nTo_);

	protected:
		void memset_(T* s, T c, size_t n);
		void memset_32(DWORD* s, DWORD c, size_t n);
		void memcpy_(T* s, const T* ct, size_t n);

	protected:
		T* m_pFrameStart, *m_pFrame;
		LONG m_lPitch;

};


template <class T, BYTE bFlags_>
CDisplayXx1<T,bFlags_>::CDisplayXx1 (BYTE* pbFrame_/*=NULL*/, LONG lPitch_/*=0*/)
{
	if ((bFlags_ & F_PALETTE) && (sizeof T == sizeof BYTE))
	{
		int nWidth  = ((VISIBLE_BLOCKS + SCREEN_BLOCKS + VISIBLE_BLOCKS) << 3) * ((bFlags_ & F_DOUBLEX) ? 2 : 1);
		int nHeight = (VISIBLE_LINES + SCREEN_LINES + VISIBLE_LINES);

		pbFrame_ = m_pbFrameStart = new BYTE [nWidth * nHeight];
		memset(m_pbFrameStart, 0, nWidth * nHeight);
		lPitch_ = nWidth;
	}

	m_pFrameStart = reinterpret_cast<T*>(pbFrame_);
	m_lPitch = lPitch_ / sizeof(T);
}


template <class T, BYTE bFlags_>
inline void CDisplayXx1<T,bFlags_>::memset_(T* s, T c, size_t n)
{
	if (sizeof T == sizeof BYTE)
		memset(s, c, n);
	else
	{
		while (n--)
			*s++ = c;
	}
}

template <class T, BYTE bFlags_>
inline void CDisplayXx1<T,bFlags_>::memset_32(DWORD* s, DWORD c, size_t n)
{
	n >>= 1;
	while (n--)
		*s++ = c;
}

template <class T, BYTE bFlags_>
inline void CDisplayXx1<T,bFlags_>::memcpy_(T* t, const T* s, size_t n)
{
	return memcpy(t, s, n*sizeof(T));
}


template <class T, BYTE bFlags_>
inline void CDisplayXx1<T,bFlags_>::LeftBorder (int nFrom_, int nTo_)
{
	// Draw the required section of the left border, if any
	if (nFrom_ < BORDER_BLOCKS && nTo_ > OFFSCREEN_BLOCKS)
	{
		int nFrom = max(OFFSCREEN_BLOCKS, nFrom_), nTo = min(nTo_, BORDER_BLOCKS);

		if (!(bFlags_ & F_YUV))
		{
			UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;
			memset_(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3)), puClut[border_col], (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
		else
		{
			int nCol = clutval[border_col];
			DWORD dwCol = (((DWORD)(awV[nCol] | awY[nCol])) << 16) | awU[nCol] | awY[nCol];
			memset_32((DWORD*)(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3))), dwCol, (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
	}
}

template <class T, BYTE bFlags_>
inline void CDisplayXx1<T,bFlags_>::RightBorder (int nFrom_, int nTo_)
{
	// Draw the required section of the right border, if any
	if (nFrom_ < (WIDTH_BLOCKS-OFFSCREEN_BLOCKS) && nTo_ > (WIDTH_BLOCKS-BORDER_BLOCKS))
	{
		int nFrom = max((WIDTH_BLOCKS-BORDER_BLOCKS), nFrom_), nTo = min(nTo_, (WIDTH_BLOCKS-OFFSCREEN_BLOCKS));

		if (!(bFlags_ & F_YUV))
		{
			UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;
			memset_(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3)), puClut[border_col], (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
		else
		{
			int nCol = clutval[border_col];
			DWORD dwCol = (((DWORD)(awV[nCol] | awY[nCol])) << 16) | awU[nCol] | awY[nCol];
			memset_32((DWORD*)(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3))), dwCol, (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
	}
}

template <class T, BYTE bFlags_>
inline void CDisplayXx1<T,bFlags_>::BorderLine (int nLine_, int nFrom_, int nTo_)
{
	m_pFrame = m_pFrameStart + (m_lPitch * nLine_);

	// Draw the required section of the border, if any
	if (nFrom_ < (WIDTH_BLOCKS-OFFSCREEN_BLOCKS) && nTo_ > OFFSCREEN_BLOCKS)
	{
		int nFrom = max(OFFSCREEN_BLOCKS, nFrom_), nTo = min(nTo_, (WIDTH_BLOCKS-OFFSCREEN_BLOCKS));

		if (!(bFlags_ & F_YUV))
		{
			UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;
			memset_(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3)), puClut[border_col], (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
		else
		{
			int nCol = clutval[border_col];
			DWORD dwCol = ((awV[nCol] | awY[nCol]) << 16) |awU[nCol] | awY[nCol];
			memset_32((DWORD*)(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3))), dwCol, (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
	}
}

template <class T, BYTE bFlags_>
inline void CDisplayXx1<T,bFlags_>::BlackLine (int nLine_, int nFrom_, int nTo_)
{
	m_pFrame = m_pFrameStart + (m_lPitch * nLine_);

	// Draw the required section of the left border, if any
	if (nFrom_ < (WIDTH_BLOCKS-OFFSCREEN_BLOCKS) && nTo_ > OFFSCREEN_BLOCKS)
	{
		int nFrom = max(OFFSCREEN_BLOCKS, nFrom_), nTo = min(nTo_, (WIDTH_BLOCKS-OFFSCREEN_BLOCKS));

		if (!(bFlags_ & F_YUV))
		{
			UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;
			memset_(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3)), 0, (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
		else
		{
			int nCol = clutval[0];
			DWORD dwCol = ((awV[nCol] | awY[nCol]) << 16) |awU[nCol] | awY[nCol];
			memset_32((DWORD*)(m_pFrame + ((nFrom-OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3))), dwCol, (nTo - nFrom) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		}
	}
}


template <class T, BYTE bFlags_>
void CDisplayXx1<T,bFlags_>::Mode1Line (int nLine_, int nFrom_, int nTo_)
{
	m_pFrame = m_pFrameStart + (m_lPitch * nLine_);
	nLine_ -= VISIBLE_LINES;

	// Draw the required section of the left border, if any
	LeftBorder(nFrom_, nTo_);

	// Draw the required section of the main screen, if any
	if (nFrom_ < (BORDER_BLOCKS+SCREEN_BLOCKS) && nTo_ > BORDER_BLOCKS)
	{
		int nFrom = max(BORDER_BLOCKS, nFrom_), nTo = min(nTo_, BORDER_BLOCKS+SCREEN_BLOCKS);
		T* pFrame = m_pFrame + ((nFrom - OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		BYTE* pbDataMem = m_pbScreenData + y_dtbl[nLine_] + (nFrom - BORDER_BLOCKS);
		BYTE* pbAttrMem = m_pbScreenData + 6144 + ((nLine_ & 0xf8) << 2) + (nFrom - BORDER_BLOCKS);
		UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;

		// The actual screen line
		for (int i = nFrom; i < nTo; i++)
		{
			BYTE bData = *pbDataMem++, bAttr = *pbAttrMem++, bInk = ATTR_FG(bAttr), bPaper = ATTR_BG(bAttr);

			// toggle the colours if we're in the inverse part of the FLASH cycle
			if (g_fFlashPhase && (bAttr & 0x80))
				swap(bInk, bPaper);

			if (!(bFlags_ & F_YUV))
			{
				T ink = puClut[bInk], paper = puClut[bPaper];

				if (!(bFlags_ & F_DOUBLEX))
				{
					pFrame[0] = (bData & 0x80) ? ink : paper;
					pFrame[1] = (bData & 0x40) ? ink : paper;
					pFrame[2] = (bData & 0x20) ? ink : paper;
					pFrame[3] = (bData & 0x10) ? ink : paper;
					pFrame[4] = (bData & 0x08) ? ink : paper;
					pFrame[5] = (bData & 0x04) ? ink : paper;
					pFrame[6] = (bData & 0x02) ? ink : paper;
					pFrame[7] = (bData & 0x01) ? ink : paper;
				}
				else
				{
					pFrame[0]  = pFrame[1]  = (bData & 0x80) ? ink : paper;
					pFrame[2]  = pFrame[3]  = (bData & 0x40) ? ink : paper;
					pFrame[4]  = pFrame[5]  = (bData & 0x20) ? ink : paper;
					pFrame[6]  = pFrame[7]  = (bData & 0x10) ? ink : paper;
					pFrame[8]  = pFrame[9]  = (bData & 0x08) ? ink : paper;
					pFrame[10] = pFrame[11] = (bData & 0x04) ? ink : paper;
					pFrame[12] = pFrame[13] = (bData & 0x02) ? ink : paper;
					pFrame[14] = pFrame[15] = (bData & 0x01) ? ink : paper;
				}
			}

			// Make sure we're using the correct depth
			else if (sizeof T == sizeof WORD)
			{
				WORD wInk = clutval[bInk], wInkUY = awU[wInk] | awY[wInk], wInkVY = awV[wInk] | awY[wInk];
				WORD wPaper = clutval[bPaper], wPaperUY = awU[wPaper] | awY[wPaper], wPaperVY = awV[wPaper] | awY[wPaper];
				WORD* pwFrame = reinterpret_cast<WORD*>(pFrame);

				if (!(bFlags_ & F_DOUBLEX))
				{
					pwFrame[0] = (bData & 0x80) ? wInkUY : wPaperUY;
					pwFrame[1] = (bData & 0x40) ? wInkVY : wPaperVY;
					pwFrame[2] = (bData & 0x20) ? wInkUY : wPaperUY;
					pwFrame[3] = (bData & 0x10) ? wInkVY : wPaperVY;
					pwFrame[4] = (bData & 0x08) ? wInkUY : wPaperUY;
					pwFrame[5] = (bData & 0x04) ? wInkVY : wPaperVY;
					pwFrame[6] = (bData & 0x02) ? wInkUY : wPaperUY;
					pwFrame[7] = (bData & 0x01) ? wInkVY : wPaperVY;
				}
				else
				{
					pwFrame[0]  = (bData & 0x80) ? wInkUY : wPaperUY;
					pwFrame[1]  = (bData & 0x80) ? wInkVY : wPaperVY;
					pwFrame[2]  = (bData & 0x40) ? wInkUY : wPaperUY;
					pwFrame[3]  = (bData & 0x40) ? wInkVY : wPaperVY;
					pwFrame[4]  = (bData & 0x20) ? wInkUY : wPaperUY;
					pwFrame[5]  = (bData & 0x20) ? wInkVY : wPaperVY;
					pwFrame[6]  = (bData & 0x10) ? wInkUY : wPaperUY;
					pwFrame[7]  = (bData & 0x10) ? wInkVY : wPaperVY;
					pwFrame[8]  = (bData & 0x08) ? wInkUY : wPaperUY;
					pwFrame[9]  = (bData & 0x08) ? wInkVY : wPaperVY;
					pwFrame[10] = (bData & 0x04) ? wInkUY : wPaperUY;
					pwFrame[11] = (bData & 0x04) ? wInkVY : wPaperVY;
					pwFrame[12] = (bData & 0x02) ? wInkUY : wPaperUY;
					pwFrame[13] = (bData & 0x02) ? wInkVY : wPaperVY;
					pwFrame[14] = (bData & 0x01) ? wInkUY : wPaperUY;
					pwFrame[15] = (bData & 0x01) ? wInkVY : wPaperVY;
				}
			}

			pFrame += (bFlags_ & F_DOUBLEX) ? 16 : 8;
		}
	}

	// Draw the required section of the right border, if any
	RightBorder(nFrom_, nTo_);
}

template <class T, BYTE bFlags_>
void CDisplayXx1<T,bFlags_>::Mode2Line (int nLine_, int nFrom_, int nTo_)
{
	m_pFrame = m_pFrameStart + (m_lPitch * nLine_);
	nLine_ -= VISIBLE_LINES;

	// Draw the required section of the left border, if any
	LeftBorder(nFrom_, nTo_);

	// Draw the required section of the main screen, if any
	if (nFrom_ < (BORDER_BLOCKS+SCREEN_BLOCKS) && nTo_ > BORDER_BLOCKS)
	{
		int nFrom = max(BORDER_BLOCKS, nFrom_), nTo = min(nTo_, BORDER_BLOCKS+SCREEN_BLOCKS);
		T* pFrame = m_pFrame + ((nFrom - OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		BYTE* pbDataMem = m_pbScreenData + (nLine_ << 5) + (nFrom - BORDER_BLOCKS);
		BYTE* pbAttrMem = pbDataMem + 0x2000;
		UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;

		// The actual screen line
		for (int i = nFrom; i < nTo; i++)
		{
			BYTE bData = *pbDataMem++, bAttr = *pbAttrMem++, bInk = ATTR_FG(bAttr), bPaper = ATTR_BG(bAttr);

			// toggle the colours if we're in the inverse part of the FLASH cycle
			if (g_fFlashPhase && (bAttr & 0x80))
				swap(bInk, bPaper);

			if (!(bFlags_ & F_YUV))
			{
				T ink = puClut[bInk], paper = puClut[bPaper];

				if (!(bFlags_ & F_DOUBLEX))
				{
					pFrame[0] = (bData & 0x80) ? ink : paper;
					pFrame[1] = (bData & 0x40) ? ink : paper;
					pFrame[2] = (bData & 0x20) ? ink : paper;
					pFrame[3] = (bData & 0x10) ? ink : paper;
					pFrame[4] = (bData & 0x08) ? ink : paper;
					pFrame[5] = (bData & 0x04) ? ink : paper;
					pFrame[6] = (bData & 0x02) ? ink : paper;
					pFrame[7] = (bData & 0x01) ? ink : paper;
				}
				else
				{
					pFrame[0]  = pFrame[1]  = (bData & 0x80) ? ink : paper;
					pFrame[2]  = pFrame[3]  = (bData & 0x40) ? ink : paper;
					pFrame[4]  = pFrame[5]  = (bData & 0x20) ? ink : paper;
					pFrame[6]  = pFrame[7]  = (bData & 0x10) ? ink : paper;
					pFrame[8]  = pFrame[9]  = (bData & 0x08) ? ink : paper;
					pFrame[10] = pFrame[11] = (bData & 0x04) ? ink : paper;
					pFrame[12] = pFrame[13] = (bData & 0x02) ? ink : paper;
					pFrame[14] = pFrame[15] = (bData & 0x01) ? ink : paper;
				}
			}

			// Make sure we're using the correct depth
			else if (sizeof T == sizeof WORD)
			{
				WORD wInk = clutval[bInk], wInkUY = awU[wInk] | awY[wInk], wInkVY = awV[wInk] | awY[wInk];
				WORD wPaper = clutval[bPaper], wPaperUY = awU[wPaper] | awY[wPaper], wPaperVY = awV[wPaper] | awY[wPaper];
				WORD* pwFrame = reinterpret_cast<WORD*>(pFrame);

				if (!(bFlags_ & F_DOUBLEX))
				{
					pwFrame[0] = (bData & 0x80) ? wInkUY : wPaperUY;
					pwFrame[1] = (bData & 0x40) ? wInkVY : wPaperVY;
					pwFrame[2] = (bData & 0x20) ? wInkUY : wPaperUY;
					pwFrame[3] = (bData & 0x10) ? wInkVY : wPaperVY;
					pwFrame[4] = (bData & 0x08) ? wInkUY : wPaperUY;
					pwFrame[5] = (bData & 0x04) ? wInkVY : wPaperVY;
					pwFrame[6] = (bData & 0x02) ? wInkUY : wPaperUY;
					pwFrame[7] = (bData & 0x01) ? wInkVY : wPaperVY;
				}
				else
				{
					pwFrame[0]  = (bData & 0x80) ? wInkUY : wPaperUY;
					pwFrame[1]  = (bData & 0x80) ? wInkVY : wPaperVY;
					pwFrame[2]  = (bData & 0x40) ? wInkUY : wPaperUY;
					pwFrame[3]  = (bData & 0x40) ? wInkVY : wPaperVY;
					pwFrame[4]  = (bData & 0x20) ? wInkUY : wPaperUY;
					pwFrame[5]  = (bData & 0x20) ? wInkVY : wPaperVY;
					pwFrame[6]  = (bData & 0x10) ? wInkUY : wPaperUY;
					pwFrame[7]  = (bData & 0x10) ? wInkVY : wPaperVY;
					pwFrame[8]  = (bData & 0x08) ? wInkUY : wPaperUY;
					pwFrame[9]  = (bData & 0x08) ? wInkVY : wPaperVY;
					pwFrame[10] = (bData & 0x04) ? wInkUY : wPaperUY;
					pwFrame[11] = (bData & 0x04) ? wInkVY : wPaperVY;
					pwFrame[12] = (bData & 0x02) ? wInkUY : wPaperUY;
					pwFrame[13] = (bData & 0x02) ? wInkVY : wPaperVY;
					pwFrame[14] = (bData & 0x01) ? wInkUY : wPaperUY;
					pwFrame[15] = (bData & 0x01) ? wInkVY : wPaperVY;
				}
			}

			pFrame += (bFlags_ & F_DOUBLEX) ? 16 : 8;
		}
	}

	// Draw the required section of the right border, if any
	RightBorder(nFrom_, nTo_);
}

template <class T, BYTE bFlags_>
void CDisplayXx1<T,bFlags_>::Mode3Line (int nLine_, int nFrom_, int nTo_)
{
	m_pFrame = m_pFrameStart + (m_lPitch * nLine_);
	nLine_ -= VISIBLE_LINES;

	// Draw the required section of the left border, if any
	LeftBorder(nFrom_, nTo_);

	// Draw the required section of the main screen, if any
	if (nFrom_ < (BORDER_BLOCKS+SCREEN_BLOCKS) && nTo_ > BORDER_BLOCKS)
	{
		int nFrom = max(BORDER_BLOCKS, nFrom_), nTo = min(nTo_, BORDER_BLOCKS+SCREEN_BLOCKS);
		T* pFrame = m_pFrame + ((nFrom - OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		BYTE* pbDataMem = m_pbScreenData + (nLine_ << 7) + ((nFrom - BORDER_BLOCKS) << 2);
		UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;

		// The actual screen line
		for (int i = nFrom; i < nTo; i++)
		{
			BYTE bData;

			if (!(bFlags_ & F_YUV))
			{
				if (!(bFlags_ & F_DOUBLEX))
				{
					bData = *pbDataMem++;
					pFrame[0] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[1] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];

					bData = *pbDataMem++;
					pFrame[2] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[3] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];

					bData = *pbDataMem++;
					pFrame[4] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[5] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];

					bData = *pbDataMem++;
					pFrame[6] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[7] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
				}
				else
				{
					bData = *pbDataMem++;
					pFrame[0] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[1] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[2] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[3] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];

					bData = *pbDataMem++;
					pFrame[4] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[5] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[6] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[7] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];

					bData = *pbDataMem++;
					pFrame[8]  = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[9]  = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[10] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[11] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];

					bData = *pbDataMem++;
					pFrame[12] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[13] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[14] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[15] = puClut[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];
				}
			}

			// Make sure we're using the correct depth
			else if (sizeof T == sizeof WORD)
			{
				ULONG ulCol;

				if (!(bFlags_ & F_DOUBLEX))
				{
					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[0] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[1] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[2] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[3] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[4] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[5] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[6] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[7] = awV[ulCol] | awY[ulCol];
				}
				else
				{
					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[0] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[1] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[2] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];
					pFrame[3] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[4] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[5] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[6] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];
					pFrame[7] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[8] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[9] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[10] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];
					pFrame[11] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0xc0) >> 6)];
					pFrame[12] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x30) >> 4)];
					pFrame[13] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x0c) >> 2)];
					pFrame[14] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[mode3_bcd48 | MD3_SWITCH((bData & 0x03)     )];
					pFrame[15] = awV[ulCol] | awY[ulCol];
				}
			}

			pFrame += (bFlags_ & F_DOUBLEX) ? 16 : 8;
		}
	}

	// Draw the required section of the right border, if any
	RightBorder(nFrom_, nTo_);
}

template <class T, BYTE bFlags_>
void CDisplayXx1<T,bFlags_>::Mode4Line (int nLine_, int nFrom_, int nTo_)
{
	m_pFrame = m_pFrameStart + (m_lPitch * nLine_);
	nLine_ -= VISIBLE_LINES;

	// Draw the required section of the left border, if any
	LeftBorder(nFrom_, nTo_);

	// Draw the required section of the main screen, if any
	if (nFrom_ < (BORDER_BLOCKS+SCREEN_BLOCKS) && nTo_ > BORDER_BLOCKS)
	{
		int nFrom = max(BORDER_BLOCKS, nFrom_), nTo = min(nTo_, BORDER_BLOCKS+SCREEN_BLOCKS);
		T* pFrame = m_pFrame + ((nFrom - OFFSCREEN_BLOCKS) << ((bFlags_ & F_DOUBLEX) ? 4 : 3));
		BYTE* pbDataMem = ((nFrom - BORDER_BLOCKS) << 2) + m_pbScreenData + (nLine_ << 7);
		UINT* puClut = (bFlags_ & F_PALETTE) ? clutval : clut;

		// The actual screen line
		for (int i = nFrom; i < nTo; i++)
		{
			BYTE bData;

			if (!(bFlags_ & F_YUV))
			{
				if (!(bFlags_ & F_DOUBLEX))
				{
					bData = *pbDataMem++;
					pFrame[0] = puClut[(bData & 0xf0) >> 4];
					pFrame[1] = puClut[ bData & 0x0f];

					bData = *pbDataMem++;
					pFrame[2] = puClut[(bData & 0xf0) >> 4];
					pFrame[3] = puClut[ bData & 0x0f];

					bData = *pbDataMem++;
					pFrame[4] = puClut[(bData & 0xf0) >> 4];
					pFrame[5] = puClut[ bData & 0x0f];

					bData = *pbDataMem++;
					pFrame[6] = puClut[(bData & 0xf0) >> 4];
					pFrame[7] = puClut[ bData & 0x0f];
				}
				else
				{
					bData = *pbDataMem++;	
					pFrame[0] = pFrame[1] = puClut[(bData & 0xf0) >> 4];
					pFrame[2] = pFrame[3] = puClut[ bData & 0x0f];

					bData = *pbDataMem++;
					pFrame[4] = pFrame[5] = puClut[(bData & 0xf0) >> 4];
					pFrame[6] = pFrame[7] = puClut[ bData & 0x0f];

					bData = *pbDataMem++;
					pFrame[8] = pFrame[9]   = puClut[(bData & 0xf0) >> 4];
					pFrame[10] = pFrame[11] = puClut[ bData & 0x0f];

					bData = *pbDataMem++;
					pFrame[12] = pFrame[13] = puClut[(bData & 0xf0) >> 4];
					pFrame[14] = pFrame[15] = puClut[ bData & 0x0f];
				}
			}

			// Make sure we're using the correct depth
			else if (sizeof T == sizeof WORD)
			{
				ULONG ulCol;

				if (!(bFlags_ & F_DOUBLEX))
				{
					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[0] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)     ];
					pFrame[1] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[2] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)     ];
					pFrame[3] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[4] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)     ];
					pFrame[5] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[6] = awU[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)     ];
					pFrame[7] = awV[ulCol] | awY[ulCol];
				}
				else
				{
					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[0] = awU[ulCol] | awY[ulCol];
					pFrame[1] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)];
					pFrame[2] = awU[ulCol] | awY[ulCol];
					pFrame[3] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[4] = awU[ulCol] | awY[ulCol];
					pFrame[5] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)];
					pFrame[6] = awU[ulCol] | awY[ulCol];
					pFrame[7] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[8] = awU[ulCol] | awY[ulCol];
					pFrame[9] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)];
					pFrame[10] = awU[ulCol] | awY[ulCol];
					pFrame[11] = awV[ulCol] | awY[ulCol];

					bData = *pbDataMem++;
					ulCol = clutval[(bData & 0xf0) >> 4];
					pFrame[12] = awU[ulCol] | awY[ulCol];
					pFrame[13] = awV[ulCol] | awY[ulCol];
					ulCol = clutval[(bData & 0x0f)];
					pFrame[14] = awU[ulCol] | awY[ulCol];
					pFrame[15] = awV[ulCol] | awY[ulCol];
				}
			}

			pFrame += (bFlags_ & F_DOUBLEX) ? 16 : 8;
		}
	}

	// Draw the required section of the right border, if any
	RightBorder(nFrom_, nTo_);
}


template <class T, BYTE bFlags_>
void CDisplayXx1<T,bFlags_>::ModeChange (int nMode_, int nLine_, int nBlock_)
{
	m_pFrame = m_pFrameStart + (m_lPitch * nLine_);
	nLine_ -= VISIBLE_LINES;

	// Make sure the mode change is on the main screen
	if (!(nLine_ >= 0 && nLine_ < SCREEN_LINES && nBlock_ >= BORDER_BLOCKS && nBlock_ < (BORDER_BLOCKS+SCREEN_BLOCKS)))
		return;

	// 4 bytes used by the ASIC to draw the next screen block
	BYTE ab[4];

	// Extract the 4 bytes from the source mode, and prepare it for the mode switch
	if (m_nMode >= 3)
	{
		BYTE* pb = m_pbScreenData + (nLine_ << 7) + ((nBlock_ - BORDER_BLOCKS) << 2);

		ab[0] = ab[1] = ((pb[0] >> 0) & 0x80) | ((pb[0] << 3) & 0x40) |
						((pb[1] >> 2) & 0x20) | ((pb[1] << 1) & 0x10) |
						((pb[2] >> 4) & 0x08) | ((pb[2] >> 1) & 0x04) |
						((pb[3] >> 6) & 0x02) | ((pb[3] << 0) & 0x01);
		ab[2] = ab[3] = pb[2];
		
	}
	else
	{
		BYTE* pData = m_pbScreenData + ((m_nMode == 1)	? y_dtbl[nLine_] + (nBlock_ - BORDER_BLOCKS)
														: (nLine_ << 5) + (nBlock_ - BORDER_BLOCKS));
		BYTE* pAttr =					(m_nMode == 1)	? m_pbScreenData + 6144 + ((nLine_ & 0xf8) << 2) + (nBlock_ - BORDER_BLOCKS)
														: pData + 0x2000;
		BYTE bData = *pData, bAttr = *pAttr;
		ab[0] = (bData & 0x77) | ((bData >> 0) & 0x80) | ((bData >> 3) & 0x08);
		ab[1] = (bData & 0x77) | ((bData << 2) & 0x80) | ((bData >> 1) & 0x08);
		ab[2] = (bAttr & 0x77) | ((bData << 4) & 0x80) | ((bData << 1) & 0x08);
		ab[3] = (bAttr & 0x77) | ((bData << 6) & 0x80) | ((bData << 3) & 0x08);
	}

	// The target mode decides how the data actually appears in the transition block
	switch (nMode_)
	{
		case 1:
		{
			BYTE* pData = m_pbScreenData + y_dtbl[nLine_] + (nBlock_ - BORDER_BLOCKS), bData = *pData;
			BYTE* pAttr = m_pbScreenData + 6144 + ((nLine_ & 0xf8) << 2) + (nBlock_ - BORDER_BLOCKS), bAttr = *pAttr;

			*pData = ab[0];
			*pAttr = ab[2];
			Mode1Line(nLine_ + VISIBLE_LINES, nBlock_, nBlock_+1);

			*pData = bData;
			*pAttr = bAttr;
			break;
		}

		case 2:
		{
			BYTE* pData = m_pbScreenData + (nLine_ << 5) + (nBlock_ - BORDER_BLOCKS), bData = *pData;
			BYTE* pAttr = pData + 0x2000, bAttr = *pAttr;

			*pData = ab[0];
			*pAttr = ab[2];
			Mode2Line(nLine_ + VISIBLE_LINES, nBlock_, nBlock_+1);

			*pData = bData;
			*pAttr = bAttr;
			break;
		}

		case 3:
		case 4:
		{
			BYTE* pb = m_pbScreenData + (nLine_ << 7) + ((nBlock_ - BORDER_BLOCKS) << 2);
			DWORD* pdw = reinterpret_cast<DWORD*>(pb), dw = *pdw;

			pb[0] = ab[0];
			pb[1] = ab[1];
			pb[2] = ab[2];
			pb[3] = ab[3];

			if (nMode_ == 3)
				Mode3Line(nLine_ + VISIBLE_LINES, nBlock_, nBlock_+1);
			else
				Mode4Line(nLine_ + VISIBLE_LINES, nBlock_, nBlock_+1);

			*pdw = dw;
			break;
		}
	}
}


#endif	// DISPLAY_H
