#include "stdafx.h"

#include "CPU.h"
#include "Memory.h"
#include "Options.h"
#include "Util.h"
#include "Mouse.h"
#include "Sound.h"
#include "Frame.h"
#include "Clock.h"
#include "Atom.h"
#include "CDrive.h"
#include "MIDI.h"
#include "Parallel.h"
#include "Serial.h"
#include "Floppy.h"

#include "Display.h"
//extern CDisplay* g_pDisplay;

extern int g_nLine;
extern int g_nLineCycle;
extern bool fContended;
extern unsigned long tstates;

extern BYTE status_reg, status_reg_active;
extern DWORD aulPalette[];	// 128 SAM colours


// Paging ports for internal and external memory
BYTE vmpr, hmpr, lmpr, lepr, hepr;
BYTE vmpr_mode, vmpr_page1, vmpr_page2;

BYTE mode3_bcd48 = 0;			// Mode 3 clut BCD bits 4 and 8 from HMPR

BYTE border, border_col;

BYTE keyboard;
BYTE status_reg, status_reg_active;
BYTE line_int;

BYTE hpen, lpen;

UINT clut[N_CLUT_REGS], clutval[N_CLUT_REGS];

BYTE keyports[9];		// 8 rows of keys (+ 1 row for unscanned keys)

// Floppy drive objects for drives 1 and 2
CDrive g_Drive1, g_Drive2;


namespace IO
{
static inline void out_vmpr(BYTE bVal_);
static inline void out_hmpr(BYTE bVal_);
static inline void out_lmpr(BYTE bVal_);
static inline void out_lepr(BYTE bVal_);
static inline void out_hepr(BYTE bVal_);

bool Init (bool fReset_/*=false*/)
{
	Exit(true);

	out_vmpr(0);					// Video in page 0, screen mode 1
	out_hmpr(0);					// Page 0 in section A, page 1 in section B, ROM0 on, ROM off
	out_lmpr(0);					// Page 0 in section C, page 1 in section D

	// Initialise external memory
	out_lepr(0);
	out_hepr(0);

	out_byte(BORDER_PORT, 0);		// Border black, speaker off

	// Floppy controller reset needed here?


	// Spectrum mode needs a bit of help to set up the paging so no pages are shared!
	if (GetOption(spectrum))
	{
		out_lmpr(LMPR_WPROT|LMPR_ROM0_OFF|0x1f);// Write-protect section A and disable ROM0, pages 31 and 0 for sections A and B
		out_hmpr(1);							// Pages 1 and 2 for sections C and D
//		out_vmpr(0);							// Screen in section B (page 0)
		out_byte(BORDER_PORT, 7);				// Border white, speaker off
	}

	// The following are initialised on powerup, but not by the reset button
	if (!fReset_)
	{
		// Set up the default palette
		for (int i = 0 ; i < N_CLUT_REGS ; i++)
			clut[i] = aulPalette[clutval[i] = (i | (i & 7) << 4)];

		// No keys pressed initially
		memset(&keyports, 0xff, sizeof keyports);
		keyboard = 0x5f;

		// // No extended keys pressed, no active (or visible) interrupts
		status_reg = status_reg_active = 0xff;

		line_int = 0xff;				// Line interrupts disabled
		hpen = lpen = 0x00;				// Clear lightpen/scan position

		// Initialise the clocks to the current date/time, the Atom disk interface and the MIDI module
		Clock::Init();
		Atom::Init();
		MIDI::Init();
		Serial::Init();
		Parallel::Init();
		Floppy::Init();

		// Insert any disks specified, or try for the default names
		static bool fFirstRun = true;
		if (fFirstRun)
		{
			g_Drive1.Insert(GetOption(disk1));
			g_Drive2.Insert(GetOption(disk2));
			fFirstRun = false;
		}
	}

	return true;
}

void Exit (bool fReinit_/*=false*/)
{
	if (!fReinit_)
	{
		g_Drive1.Eject();
		g_Drive2.Eject();

		Floppy::Exit();
		Parallel::Exit();
		Serial::Exit();
		MIDI::Exit();
		Atom::Exit();
		Clock::Exit();
	}
}



static inline void out_lepr(BYTE val)
{
	lepr = val;

	if (HMPR_MCNTRL)
	{
		int ExternalPage = N_PAGES_MAIN+val;
		PageIn(SECTION_C, ExternalPage);
	}
}

static inline void out_hepr (BYTE val)
{
	hepr = val;

	if (HMPR_MCNTRL)
	{
		int ExternalPage = N_PAGES_MAIN+val;
		PageIn(SECTION_D, ExternalPage);
	}
}

static inline void out_vmpr (BYTE val)
{
	vmpr = (val & 0x7f);

	// Set the screen mode (1 to 4) - the ASIC changed mode before page, so this must be done first
	vmpr_mode = vmpr & VMPR_MODE_MASK;
	g_pDisplay->SetMode((vmpr_mode >> 5) + 1, g_nLine, g_nLineCycle);

	// Extract the page number(s) for faster access by the memory writing functions
	vmpr_page1 = VMPR_SCREEN_PAGE;
	vmpr_page2 = (vmpr_page1+1) & 0x1f;
}

static inline void out_hmpr (BYTE val)
{
	// Update the HMPR
	hmpr = val;

	// Update BCD bits 4 and 8 of the clut that is used by Sam mode 3
	mode3_bcd48 = (hmpr & 0x60) >> 3;

	// Put the relevant RAM bank into section C
	PageIn(SECTION_C, HMPR_PAGE);

	// Put either RAM or ROM1 into section D
	if (lmpr & LMPR_ROM1)
		PageIn(SECTION_D, ROM1);
	else
		PageIn(SECTION_D,(HMPR_PAGE + 1)& 0x1f);

	// Perform changes to external memory
	if (HMPR_MCNTRL)
	{
		out_hepr(hepr);
		out_lepr(lepr);
	}
}


static inline void out_lmpr(BYTE val)
{
	lmpr = val;

	// Put either RAM or ROM 0 into section A
	if (lmpr & LMPR_ROM0_OFF)
		PageIn(SECTION_A, LMPR_PAGE);
	else
		PageIn(SECTION_A, ROM0);

	// Put the relevant bank into section B
	PageIn(SECTION_B, (LMPR_PAGE + 1) & 0x1f);

	// Put either RAM or ROM 1 into section D
	if (lmpr & LMPR_ROM1)
		PageIn(SECTION_D, ROM1);
	else
		PageIn(SECTION_D, (HMPR_PAGE + 1) &0x1f);
}

static inline void out_clut(WORD wPort_, BYTE bVal_)
{
	wPort_ &= (N_CLUT_REGS-1);			// 16 clut registers, so only the bottom 4 bits are significant
	bVal_ &= (N_PALETTE_COLOURS-1);		// 128 colours, so only the bottom 7 bits are significant

	if (clut[wPort_] != aulPalette[bVal_])
	{
		g_pDisplay->Update(g_nLine, g_nLineCycle);

		clut[wPort_] = (DWORD)aulPalette[clutval[wPort_] = bVal_];
	}
}


BYTE In (WORD wPort_)
{
	BYTE bPortLow = wPort_ & 0xff;

	switch (bPortLow)
	{
		// keyboard 1 / mouse
		case KEYBOARD_PORT:
		{
			int highbyte = wPort_ >> 8, res = 0x1f;

			if (highbyte == 0xff)
			{
				res = keyports[8] & 0x1f;

				if (GetOption(mouse) && res == 0x1f)
					res = Mouse::Read(g_dwCycleCounter) & 0x1f;
			}
			else
			{
				if (!(highbyte & 0x80)) res &= keyports[7] & 0x1f;
				if (!(highbyte & 0x40)) res &= keyports[6] & 0x1f;
				if (!(highbyte & 0x20)) res &= keyports[5] & 0x1f;
				if (!(highbyte & 0x10)) res &= keyports[4] & 0x1f;
				if (!(highbyte & 0x08)) res &= keyports[3] & 0x1f;
				if (!(highbyte & 0x04))
				{
					res &= keyports[2] & 0x1f;

					// Once the keyboard is being read we must have finished booted, so unpatch the ROM
					Memory::FastStartPatch(false);

					static int nFastBoot = 200 * GetOption(autoboot);
					if (nFastBoot)
					{
						if (--nFastBoot)
							keyports[2] &= ~(1 << 7);	// Press F9
						else
							keyports[2] |= (1 << 7);	// Release F9
					}
				}

				if (!(highbyte & 0x02)) res &= keyports[1] & 0x1f;
				if (!(highbyte & 0x01)) res &= keyports[0] & 0x1f;
			}

/*
			// Play with ear bit to simulate loading edges!
			static int i = 0;
			if (!(++i %= 60))
			{
				keyboard ^= 0x40;

//				if (GetOption(loadsound))
					Sound::Click((keyboard & 0x40) >> 2);
			}
*/

			return (keyboard & 0xe0) | res;
		}

		// keyboard 2
		case STATUS_PORT:
		{
			int highbyte = wPort_ >> 8, res = 0xe0;

			if (!(highbyte & 0x80)) res &= keyports[7] & 0xe0;
			if (!(highbyte & 0x40)) res &= keyports[6] & 0xe0;
			if (!(highbyte & 0x20)) res &= keyports[5] & 0xe0;
			if (!(highbyte & 0x10)) res &= keyports[4] & 0xe0;
			if (!(highbyte & 0x08)) res &= keyports[3] & 0xe0;
			if (!(highbyte & 0x04)) res &= keyports[2] & 0xe0;
			if (!(highbyte & 0x02)) res &= keyports[1] & 0xe0;
			if (!(highbyte & 0x01)) res &= keyports[0] & 0xe0;

			return (res & 0xe0) | (status_reg & 0x1f);
		}


		// Banked memory management
		case VMPR_PORT:		return vmpr | 0x80;
		case HMPR_PORT:		return hmpr;
		case LMPR_PORT:		return lmpr;

		// SAMBUS and DALLAS clock ports
		case CLOCK_PORT:	return Clock::In(wPort_);

		// HPEN and LPEN ports
		case LPEN_PORT:
			if ((wPort_ & 0x1ff) == HPEN_PORT)
				return g_pDisplay->GetHPen(g_nLine);
			else
				return lpen;

		// Spectrum ATTR port)
		case ATTR_PORT:		return  g_pDisplay->GetAttr();

		// External memory
		case HEPR_PORT:		return hepr;
		case LEPR_PORT:		return lepr;

		// Parallel ports
		case PRINTL1_STAT:
		case PRINTL2_STAT:
		case PRINTL1_DATA:
		case PRINTL2_DATA:
			return Parallel::In(wPort_);

		// Serial ports
		case SERIAL1:
		case SERIAL2:
			return Serial::In(wPort_);

		// MIDI IN/Network
		case MIDI_PORT:
			return MIDI::In();

		default:
		{
			// Dunno if there's a special reason these return zero!  Anyone?
			if (bPortLow < 0x10)
				return 0;

			// Floppy drive 1
			else if ((wPort_ & 0xf8) == FLOPPY1_BASE)
			{
				// Read from floppy drive 1, if present
				if (GetOption(drive1))
					return g_Drive1.In(wPort_);
			}

			// Floppy drive 2 *OR* the ATOM hard disk
			else if ((wPort_ & 0xf8) == FLOPPY2_BASE)
			{
				// Read from floppy drive 2, if present...
				if (GetOption(drive2) == 1)
					return g_Drive2.In(wPort_);

				// ... or Atom hard disk
				else if (GetOption(drive2) == 2)
					return Atom::In(wPort_);
			}

			// Only unsupported hardware should reach here
			else
				TRACE("*** Unhandled read: %#04x (%d)\n", wPort_, wPort_&0xff);
		}
	}

	// Default to returning 0xff
	return 0xff;
}


// The actual port input and output routines
void Out (WORD wPort_, BYTE bVal_)
{
	BYTE bPortLow = wPort_ & 0xff;

	// border colour and (possibly) speaker control
	switch (bPortLow)
	{
		case BORDER_PORT:
		{
			bool fScreenOffChange = ((border ^ bVal_) & 0x80) && ((vmpr_mode >= MODE_3));

			// Has the border colour has changed colour or the screen been enabled/disabled?
			if (BORD_COL(bVal_) != border_col || fScreenOffChange)
				g_pDisplay->Update(g_nLine, g_nLineCycle);

//			if (GetOption(savesound))
//				bVal_ ^= ((~bVal_ & 0x08) << 1);

			// If the speaker bit has been toggled, generate a click
			if ((border ^ bVal_) & 0x10)
				Sound::Click(bVal_ & 0x10);

			// Store the new border value and extract the border colour for faster access by the video routines
			border = bVal_;
			border_col = BORD_VAL(bVal_);
			keyboard = (keyboard & 0x7f) | BORD_SOFF;

			// If the screen state has changed, we need to reconsider memory contention changes
			if (fScreenOffChange)
				CPU::UpdateContention();
		}
		break;

		// Banked memory management
		case VMPR_PORT:
		{
			// Changes to screen mode and screen page are visible at different times

			// Has the screen mode changed?
			if ((bVal_ & VMPR_MODE_MASK) != vmpr_mode)
			{
				// Are either the current mode or the new mode 3 or 4?  i.e. bit MDE1 is set
				if ((bVal_ | vmpr) & 0x40)
				{
					// Changes to the screen MODE are visible straight away
					g_pDisplay->Update(g_nLine, g_nLineCycle);

					// Change only the screen MODE for the transition block
					out_vmpr((bVal_ & VMPR_MODE_MASK) | (vmpr & ~VMPR_MODE_MASK));
				}
				// Otherwise both modes are 1 or 2
				else
				{
					// There are no visible changes in the transition block
					g_pDisplay->Update(g_nLine, g_nLineCycle + VIDEO_DELAY);

					// Do the whole change here - the check below will not be triggered
					out_vmpr(bVal_);
				}

				// Video mode change may have affected memory contention
				CPU::UpdateContention();
			}

			// Has the screen page changed?
			if (vmpr_page1 != (bVal_ & 0x1f))
			{
				// Changes to screen PAGE aren't visible until 8 tstates later
				// as the memory has been read by the ASIC already
				g_pDisplay->Update(g_nLine, g_nLineCycle + VIDEO_DELAY);

				out_vmpr(bVal_);
			}
		}
		break;

		case HMPR_PORT:
			if (hmpr != bVal_)
			{
				// If we are in mode 3 and the BCD4/8 values have changed then we need to update the screen
				if ((vmpr_mode == MODE_3) && (((bVal_ & 0x60) >> 3) != mode3_bcd48))
					g_pDisplay->Update(g_nLine, g_nLineCycle);

				out_hmpr(bVal_);
			}
			break;

		case LMPR_PORT:
			if (lmpr != bVal_)
				out_lmpr(bVal_);
			break;

		// SAMBUS and DALLAS clock ports
		case CLOCK_PORT:
			Clock::Out(wPort_, bVal_);
			break;

		case CLUT_BASE_PORT:
			out_clut(wPort_ >> 8, bVal_);
			break;

		// External memory
		case HEPR_PORT:
			out_hepr(bVal_);
			break;

		case LEPR_PORT:
			out_lepr(bVal_);
			break;

		case LINE_PORT:
			line_int = bVal_;
			break;

		case SOUND_DATA:
			Sound::Out(wPort_, bVal_);
			break;

		// Parallel ports 1 and 2
		case PRINTL1_STAT:
		case PRINTL1_DATA:
		case PRINTL2_STAT:
		case PRINTL2_DATA:
		case 0x7f:
			Parallel::Out(wPort_, bVal_);
			break;

		// Serial ports 1 and 2
		case SERIAL1:
		case SERIAL2:
			Serial::Out(wPort_, bVal_);
			break;

		// MIDI OUT/Network
		case MIDI_PORT:
		{
			// Only transmit a new byte if one isn't already being sent
			if (!(lpen & LPEN_TXFMST))
			{
				// Set the TXFMST bit in LPEN to show that we're transmitting something
				lpen |= LPEN_TXFMST;

				// Output the byte using the platform specific implementation
				MIDI::Out(bVal_);
			}
			break;
		}

		default:
			// Floppy drive 1
			if ((wPort_ & FLOPPY_MASK) == FLOPPY1_BASE)
			{
				// Write to floppy drive 1, if present
				if (GetOption(drive1))
					g_Drive1.Out(wPort_, bVal_);
			}

			// Floppy drive 2 *OR* the ATOM hard disk
			else if ((wPort_ & FLOPPY_MASK) == FLOPPY2_BASE)
			{
				// Write to floppy drive 2, if present...
				if (GetOption(drive2) == 1)
					g_Drive2.Out(wPort_, bVal_);

				// ... or Atom hard disk
				else if (GetOption(drive2) == 2)
					Atom::Out(wPort_,bVal_);
			}

			// Only unsupported hardware should reach here
			else
				TRACE("*** Unhandled write: %#04x (LSB=%d), %#02x (%d)\n", wPort_, bPortLow, bVal_, bVal_);
	}
}

};
