#include "stdafx.h"

#include "CPU.h"
#include "Memory.h"
#include "Util.h"
#include "Options.h"

////////////////////////////////////////////////////////////////////////////////

// Single block holding all memory needed
BYTE* pMemory;

// Master read and write lists that are static for a given memory configuration
BYTE* apbPageReadPtrs[TOTAL_PAGES];
BYTE* apbPageWritePtrs[TOTAL_PAGES];

// Page numbers present in each of the 4 sections in the 64K address range
int anSectionPages[4];

// Array of pointers for memory to use when reading from or writing to each each section
BYTE* apbSectionReadPtrs[4];
BYTE* apbSectionWritePtrs[4];

////////////////////////////////////////////////////////////////////////////////

namespace Memory
{
static int ReadRoms ();

// Allocate and initialise memory
bool Init ()
{
	Exit(true);

	// Allocate and clear all we're likely to need in a single block (512K + 4MB external + 2 ROMs + 2 scratch)
	if (!pMemory && !(pMemory = new BYTE[TOTAL_PAGES * MEM_PAGE_SIZE]))
	{
		Message(msgError, "Out of memory!");
		return false;
	}

	// Clear out memory - allows use of the fast-boot option to skip SAM's memory check and initialisation
	memset(pMemory, 0, TOTAL_PAGES * MEM_PAGE_SIZE);

	// Start with all memory available, and fully readable and writable
	for (int i = 0 ; i < TOTAL_PAGES ; i++)
		apbPageReadPtrs[i] = apbPageWritePtrs[i] = pMemory + (i * MEM_PAGE_SIZE);

	// Read the ROMs and set the memory as read-only
	ReadRoms();
	apbPageWritePtrs[ROM0] = apbPageWritePtrs[ROM1] = apbPageWritePtrs[SCRATCH_WRITE];

	// Are we in 256K mode?
	if (GetOption(mainmem) == 256)
	{
		// Invalidate the top 256K of main memory
		for (int i = N_PAGES_MAIN/2 ; i < N_PAGES_MAIN ; i++)
			apbPageReadPtrs[i] = (apbPageWritePtrs[i] = apbPageWritePtrs[SCRATCH_WRITE]) - MEM_PAGE_SIZE;
	}

	// Make any absent portions of external memory read-only (note: external memory is added top-down!)
	int nExternal = min(GetOption(externalmem), MAX_EXTERNAL_MB);
	for (int nExt = 0 ; nExt < (MAX_EXTERNAL_MB - nExternal) * N_PAGES_1MB ; nExt++)
		apbPageWritePtrs[N_PAGES_MAIN + nExt] = apbPageWritePtrs[SCRATCH_WRITE];

	return true;
}

void Exit (bool fReinit_/*=false*/)
{
	if (!fReinit_)
	{
		if (pMemory) { delete pMemory; pMemory = NULL; }
	}
}

// Patch ROM0 and ROM1 to speed up the initial boot 
void FastStartPatch (bool fPatch_/*=true*/)
{
	static bool fPatched = false;

	// Check the option is enabled, and for ROM version 3.x before changing anything, just in case
	if ((fPatched != fPatch_) && apbPageReadPtrs[ROM0][0x000f] / 10 == 3)
	{
		if (fPatched = fPatch_)
		{
			apbPageReadPtrs[ROM0][0x00b5] = 0x06;	// JR NZ,e -> LD B,e    to short-circuit BC delay loop
			apbPageReadPtrs[ROM0][0x0796] = 0x06;	// JR NZ,e -> LD B,e    to short-circuit BC delay loop
			apbPageReadPtrs[ROM1][0x2bc9] = 0x06;	// LDIR    -> LD B,176  to avoid page clear slowdown
			apbPageReadPtrs[ROM1][0x2bd0] = 0x01;	// LD E,64 -> LD E,1    only one read/write test per page
		}
		else
		{
			// Reverse what we did above
			apbPageReadPtrs[ROM0][0x00b5] = 0x20;
			apbPageReadPtrs[ROM0][0x0796] = 0x10;
			apbPageReadPtrs[ROM1][0x2bc9] = 0xed;
			apbPageReadPtrs[ROM1][0x2bd0] = 0x40;
		}
	}
}

// Read the ROM image into the ROM area of our paged memory
static bool LoadRomImage (const char* pcszImage_, int nPage_)
{
	FILE* file;

	if (!(file = fopen(pcszImage_, "rb")))
		Message(msgError, "Failed to open ROM image: %s", pcszImage_);
	else
	{
		// Read the image, warning if the image is not exactly 16K
		if (fread(apbPageReadPtrs[nPage_], 1, MEM_PAGE_SIZE, file) != MEM_PAGE_SIZE)
			Message(msgWarning, "ROM image (%s) is not 16K!", pcszImage_);

		fclose(file);
		return true;
	}

	return false;
}

int ReadRoms ()
{
	fprintf(stderr, "ROM0: %s\n", GetOption(rom0));

	bool f = LoadRomImage(GetOption(rom0), ROM0) && LoadRomImage(GetOption(rom1), ROM1);

	// The fast reset option requires a few temporary ROM patches that are reversed later
	if (GetOption(fastreset))
		FastStartPatch();

	// If Spectrum mode is enabled, load the Spectrum ROM into RAM page 31
	if (GetOption(spectrum))
		f &= LoadRomImage("spectrum.rom", 31);

	return f;
}

};	// namespace Memory
