#ifndef MEMORY_H
#define MEMORY_H

#include "SAM.h"
#include "CPU.h"
#include "Frame.h"
#include "Video.h"
#include "IO.h"

extern int g_nLastLineNo;

namespace Memory
{
	bool Init ();
	void Exit (bool fReinit_=false);

	void FastStartPatch (bool fPatch_=true);
}

enum eSection { SECTION_A, SECTION_B, SECTION_C, SECTION_D };

// The ROMs and scratch memory go after the regular RAM
enum { ROM0 = (N_PAGES_MAIN + (N_PAGES_1MB * MAX_EXTERNAL_MB)), ROM1, SCRATCH_READ, SCRATCH_WRITE, TOTAL_PAGES };

extern unsigned char dtbl_y[6144];

extern int anSectionPages[4];
extern BYTE* apbSectionReadPtrs[4];
extern BYTE* apbSectionWritePtrs[4];


// Map a 16-bit address through the memory indirection - allows fast paging
#define VPAGE(addr) ((addr) >> 14)
#define RPAGE(addr) (anSectionPages[VPAGE(addr)])
#define MEM_R(addr) (apbSectionReadPtrs[VPAGE(addr)][(addr) & (MEM_PAGE_SIZE-1)])
#define MEM_W(addr) (apbSectionWritePtrs[VPAGE(addr)][(addr) & (MEM_PAGE_SIZE-1)])


void write_to_screen_vmpr0 (WORD wAddr_);
void write_to_screen_vmpr1 (WORD wAddr_);
void write_word (WORD wAddr_, WORD wVal_);

inline BYTE read_byte(WORD addr)
{
	return MEM_R(addr);
}

inline WORD read_word(WORD addr)
{
	return MEM_R(addr) | (MEM_R(addr+1) << 8);
}

inline void write_byte (WORD wAddr_, BYTE bVal_)
{
	// Does the write fall within the first display page?
	if (RPAGE(wAddr_) == vmpr_page1)
		write_to_screen_vmpr0(wAddr_);

	// Does the write fall within the second display page? (modes 3 and 4 only)
	else if ((RPAGE(wAddr_) == vmpr_page2) && (vmpr_mode > MODE_2))
		write_to_screen_vmpr1(wAddr_);

	// Write the byte to memory
	MEM_W(wAddr_) = bVal_;
}

inline void write_word(WORD wAddr_, WORD wVal_)
{
	// Write the low byte then the high byte
	write_byte(wAddr_, wVal_ & 0xff);
	write_byte(wAddr_+1, wVal_ >> 8);
}


// Page in real memory page at <nSection_>, where <nSection_> is in range 0..3
inline void PageIn (eSection nSection_, int nPage_)
{
	// Remember the page that's now occupying the section
	anSectionPages[nSection_] = nPage_;

	// Look up the relevant read and write pointers for the section
	apbSectionReadPtrs[nSection_] = apbPageReadPtrs[nPage_];
	apbSectionWritePtrs[nSection_] = apbPageWritePtrs[nPage_];

	// Check for write protected RAM in section A
	if ((nSection_ == SECTION_A) && (lmpr & LMPR_WPROT))
		apbSectionWritePtrs[nSection_] = apbPageWritePtrs[SCRATCH_WRITE];
}


inline void write_to_screen_vmpr0 (WORD wAddr_)
{
	// Limit the address to the 16K page we're considering
	wAddr_ &= 0x3fff;

	// The display area affected by the write depends on the current screen mode
	int nLine1, nLine2;
    switch (vmpr_mode)
	{
		case MODE_1:
			// If writing to the main screen data, invalidate the screen line we're writing to
			if (wAddr_ < 6144)
				nLine1 = nLine2 = dtbl_y[wAddr_];

			// If writing to the attribute area, invalidate the 8 lines containing the attribute
			else if (wAddr_ < 6912)
			{
				nLine1 = ((wAddr_-6144) & 0xffe0) >> 2;
				nLine2 = nLine1 + 7;
			}

			// Doesn't fall within mode 1 screen
			else
				return;

			break;

		case MODE_2:
			// If the write falls within the screen data or attributes, invalidate the line
			if (wAddr_ < 6144 || (wAddr_ >= 8192 && wAddr_ < (8192+6144)))
			{
				nLine1 = nLine2 = (wAddr_ & 0x1fff) >> 5;
				break;
			}
			return;

		// Modes 3 and 4
		default:
		{
			// The write is to the first 16K of a mode 3 or 4 screen
			nLine1 = nLine2 = (wAddr_ >> 7);
			break;
		}
	}

	nLine1 += TOP_BORDER_LINES;
	nLine2 += TOP_BORDER_LINES;

	if (nLine2 >= g_nLastLineNo && nLine1 <= g_nLine)
		g_pDisplay->Update(g_nLine, g_nLineCycle);
}  

inline void write_to_screen_vmpr1 (WORD wAddr_)
{
	// Limit the address to the 16K page we're considering
	wAddr_ &= 0x3fff;

	if (wAddr_ < 8192)
	{
		int nLine = TOP_BORDER_LINES + ((wAddr_ + MEM_PAGE_SIZE) >> 7);

		if (nLine >= g_nLastLineNo && nLine <= g_nLine)
			g_pDisplay->Update(g_nLine, g_nLineCycle);
	}
}


#endif	// MEMORY_H
