#include "stdafx.h"

#include "CPU.h"
#include "Memory.h"
#include "Util.h"
#include "Options.h"
#include "UI.h"

const char* const MSGBOX_TITLE = "SimCoupe";

static FILE *logfp = NULL;


// Report an info, warning, error or fatal message.  Exit if a fatal message has been reported
void Message (eMsgType eType_, const char* pcszFormat_, ...)
{
	va_list args;
	va_start(args, pcszFormat_);

	const char* pcszType_ = "";
	switch (eType_)
	{
		case msgWarning:	pcszType_ = "warning: ";	break;
		case msgError:		pcszType_ = "error: ";		break;
		case msgInfo:		pcszType_ = "info: ";		break;
		case msgFatal:		pcszType_ = "fatal: ";		break;
	}

	char szMessage[512];
	strcpy(szMessage, pcszType_);
	char* pszMessage = szMessage + strlen(szMessage);
	vsprintf(pszMessage, pcszFormat_, args);
	strcat(szMessage, "\n");

	// Write to the debugger
	TRACE(szMessage);

	// Write to file, if open
	if (logfp)
		fputs(szMessage, logfp);

	UI::ShowMessage(eType_, pszMessage);
}

// Log file utilities
int open_log ()
{
	if (*GetOption(logfile) && (logfp = fopen(GetOption(logfile), "w")) == NULL) {
		Message(msgError, "can't open log file");
		return -1;
	}	
	return 0;
} 

void close_log ()
{
	if (logfp)
		fclose(logfp);	
}

////////////////////////////////////////////////////////////////////////////////

// Constants
#define TRACE_BUFFER_SIZE		2048

__int64 g_llLastTrace, g_llFrequency;
bool fSkipTimer = FALSE;
bool g_fTraceEnabled = TRUE;;


#ifdef _DEBUG

static void __cdecl TraceOutputString (const char *pcszFormat, const char *pcvArgs);
static void OutputTimeString ();


// Output a formatted debug message
void __cdecl TraceOutputString (const char *pcszFormat, ...)
{
	// Get a pointer to the arguments
	va_list	pcvArgs;
	va_start (pcvArgs, pcszFormat);

	TraceOutputString(pcszFormat, pcvArgs);

	va_end(pcvArgs);
}


// Output a formatted debug message
static void __cdecl TraceOutputString (const char *pcszFormat, const char *pcvArgs)
{
	char szTrace[TRACE_BUFFER_SIZE];

	if (!g_fTraceEnabled)
		return;

	// If the format string doesn't require any parameters, don't bother formatting it
	if (!strchr(pcszFormat, '%'))
		strcpy(szTrace, pcszFormat);

	// Format the string to a buffer, verify that it doesn't overflow and corrupt memory
	else
		vsprintf(szTrace, pcszFormat, (va_list)pcvArgs);

	// Output the time string (if available) and the debug message
	OutputTimeString();
	OutputDebugString (szTrace);

	// Flag to skip the timer on the next trace if the current message doesn't end in a newline
	fSkipTimer = (szTrace[strlen(szTrace)-1] != '\n');
}


// Output a formatted debug message in hex blocks with ASCII
void TraceOutputString (const BYTE *pcb, UINT uLen/*=0*/, bool fASCII/*=TRUE*/, BOOL fNewline/*=TRUE*/)
{	
	static BYTE abHexBytes[] = "0123456789ABCDEF";
	char szTrace[TRACE_BUFFER_SIZE];

	if (!g_fTraceEnabled)
		return;

	// If the length wasn't given or was zero, assume it's a string and use strlen to get the length
	if (!uLen)
		uLen = strlen(reinterpret_cast<const char *>(pcb));

	// Loop while there is still data to process
	while (uLen > 0)
	{
		BYTE *pabLinePos = reinterpret_cast<BYTE*>(szTrace);
		BYTE *pabASCIIPos = (BYTE*)szTrace + 16*3+4;

		memset(pabLinePos, ' ', 80);

		// Append each hex byte until no more bytes or this line is full.
		for (int i = 0 ; uLen && i < 16 ; i++, pcb++, uLen--)
		{
			// Make the data more readable by adding an extra space every 4 bytes
			if (i && !(i % 4))
				*pabLinePos++ = ' ';

			// Store the hex byte
			*pabLinePos++ = abHexBytes[*pcb >> 4];
			*pabLinePos++ = abHexBytes[*pcb & 15];
			*pabLinePos++ = ' ';

			// If ASCII is enabled, store the ASCII character or '.' for out-of-range characters
			if (fASCII)
			{
				*pabASCIIPos++ = (*pcb > 0x1F && *pcb < 0x7F) ? *pcb : '.';
				if (*pcb == '%')
					*pabASCIIPos++ = '%';
			}
		}
						
		// Terminate with a newline and a null
		if (fASCII)
			pabLinePos = pabASCIIPos;
		if (fNewline)
			*pabLinePos++ = '\n';
		*pabLinePos   = '\0';

		// Output the time string (if available) and the debug message
		OutputTimeString ();
		OutputDebugString (szTrace);

		// Flag to skip the timer on the next trace if the current message doesn't end in a newline
		fSkipTimer = (szTrace[strlen(szTrace)-1] != '\n');
	}
}


// Output the time string to the debug device
void OutputTimeString ()
{
	// Return if we're to skip this time
#ifdef _WIN32
	if (fSkipTimer)
#endif
	{
		fSkipTimer = FALSE;
		return;
	}

#ifdef _WIN32	

	// Fetch the current system time and convert to a file time
	__int64 llCurrent;
	QueryPerformanceCounter(reinterpret_cast<LARGE_INTEGER*>(&llCurrent));

	// Work out how much time has elapsed since the trace counter was reset
	__int64 llElapsed;
	if (g_llLastTrace)
		llElapsed = llCurrent - g_llLastTrace;

	// The trace counter was not set, so set its initial value
	else
	{
		// Get the number 
		QueryPerformanceFrequency(reinterpret_cast<LARGE_INTEGER*>(&g_llFrequency));
		g_llFrequency /= 1000;

		g_llLastTrace = llCurrent;
		llElapsed = 0;
	}

	// Convert the 100ns value to a ms value
	DWORD dwElapsed = static_cast<DWORD>(llElapsed / g_llFrequency);

	// Break the elapsed time into seconds, minutes and milliseconds
	DWORD dwMilisecs = dwElapsed % 1000;
	DWORD dwSecs = (dwElapsed /= 1000) % 60;
	DWORD dwMins = (dwElapsed /= 60) % 100;

	// Form the time string and send to the debugger
	char szTime [16];
	wsprintf (szTime, "%02lu:%02lu.%03lu  ", dwMins, dwSecs, dwMilisecs);
	OutputDebugString (szTime);

#endif	// _WIN32
}

#else

void __cdecl TraceOutputString (const char *, ...) { }

#endif
