//	Notes:
//		The worst part of this is the on-screen profile stats, as they're
//		drawn using GDI functions every frame, but only calculated every
//		1 second.  Majorly inefficient and gives a speed drop when enabled!
//		Will probably add an extra video surface to hold them to speed up
//		blitting, and only redraw it on the surface when they change.
//
//		The WaitVBlank() vsync method needs additional testing on when it can
//		be used, and probably how many to do for a perfect sync.  Only really
//		useful for 100Hz or 150Hz monitor refresh rates.



#include "stdafx.h"

#include "Frame.h"
#include "Display.h"
#include "CPU.h"
#include "Util.h"
#include "Options.h"
#include "Video.h"
#include "CDrive.h"
#include "Profile.h"

extern HWND g_hwnd;
extern DWORD g_dwColourKey;

bool fDrawFrame = true;

// SAM palette colours to use for the floppy drive LED on/off
#define LED_OFF_COLOUR			8
#define LED_ON_COLOUR			70


namespace Frame
{
static int nFrame, nDrawnFrames;	// Frame count in last second and number of those actually drawn
static DWORD dwStartTime, dwLastTime;

static HANDLE g_hEvent;
static HFONT g_hfontStatus;
static MMRESULT g_hTimer;
static char g_szStatus[128], g_szProfile[128];

// Timer handler, called every 20ms
void CALLBACK TimeCallback (UINT uTimerID_, UINT uMsg_, DWORD dwUser_, DWORD dw1_, DWORD dw2_)
{
	// Signal that the next frame is due
	SetEvent(g_hEvent);

	// We could do with updating the input in here to give a better keyboard response
	// for slow machines, but DirectInput does nasty things if we try!
}


bool Init ()
{
	// Clear the frame buffer surfaces
	Clear();
	Video::UpdateOverlay();

	// Create an event that will be set every 20ms for the 50Hz sync
	if (!(g_hEvent = CreateEvent(NULL, TRUE, TRUE, NULL)))
		Message(msgWarning, "Failed to create sync event object (%#08lx)", GetLastError());

	// Set a timer to fire every every 20ms for our 50Hz frame synchronisation
	else if (!(g_hTimer = timeSetEvent(1000/EMULATED_FRAMES_PER_SECOND, 0, TimeCallback, 0, TIME_PERIODIC|TIME_CALLBACK_FUNCTION)))
		Message(msgWarning, "Failed to start sync timer (%#08lx)", GetLastError());
	else
	{
		// Create an 8pt font to draw the stats on the back surface
		LOGFONT lf;
		ZeroMemory(&lf, sizeof lf);
		lf.lfCharSet= DEFAULT_CHARSET;
		lf.lfHeight = -8;
		g_hfontStatus = CreateFontIndirect(&lf);

		// Success :-)
		return true;
	}

	// Clean up and return failure
	Exit();
	return false;
}

void Exit ()
{
	if (g_hfontStatus) { DeleteObject(g_hfontStatus); g_hfontStatus = NULL; }

	if (g_hEvent)	{ CloseHandle(g_hEvent); g_hEvent = NULL; }
	if (g_hTimer)	{ timeKillEvent(g_hTimer); g_hTimer = NULL; }
	if (g_pDisplay)	{ delete g_pDisplay; g_pDisplay = NULL; }
}

void Start ()
{
	dwStartTime = timeGetTime();
}


void End ()
{
	// If we've done a seconds worth we need to update the FPS counter
	DWORD dwNow = timeGetTime(), dwDiff = dwNow - dwLastTime;
	if (dwDiff >= 1000)
	{
		sprintf(g_szProfile, "%d%%, %dfps %s", nFrame * 2, nDrawnFrames, Profile::GetStats());

		if (!GetOption(fullscreen))
			SetWindowText(g_hwnd, GetOption(profile) ? g_szProfile : "SimCoupe/Win32");

		// Reset the counters to wait for the next second
		dwLastTime = dwNow;
		nFrame = nDrawnFrames = 0;
		Profile::Reset();
	}

	// If the current frame is being draw, update the image and counter
	if (fDrawFrame)
	{
		g_pDisplay->Update();
		nDrawnFrames++;
	}

	nFrame++;


	ProfileStart(Blit);

	// If enabled, draw the floppy LEDs in the bottom-left of the display
	if (pddsBack && pddsFront && GetOption(floppyleds))
	{
		DDSURFACEDESC ddsdBack, ddsdFront;
		ddsdBack.dwSize = ddsdFront.dwSize = sizeof ddsdBack;
		pddsBack->GetSurfaceDesc(&ddsdBack);
		pddsFront->GetSurfaceDesc(&ddsdFront);

		RECT rBack = { 0, 0, ddsdBack.dwWidth, ddsdBack.dwHeight };

		RECT rLed = { rBack.left + GetOption(sourcesize), rBack.bottom - 4, rBack.left + (7 * GetOption(sourcesize)), rBack.bottom - 2 };

		DWORD dwOn  = (ddsdFront.ddpfPixelFormat.dwFlags & DDPF_FOURCC) ? (((DWORD)(awV[LED_ON_COLOUR] | awY[LED_ON_COLOUR])) << 16) | awU[LED_ON_COLOUR] | awY[LED_ON_COLOUR] : aulPalette[LED_ON_COLOUR];
		DWORD dwOff = (ddsdFront.ddpfPixelFormat.dwFlags & DDPF_FOURCC) ? (((DWORD)(awV[LED_OFF_COLOUR] | awY[LED_OFF_COLOUR])) << 16) | awU[LED_OFF_COLOUR] | awY[LED_OFF_COLOUR] : aulPalette[LED_OFF_COLOUR];

		DDBLTFX bltfx;
		bltfx.dwSize = sizeof bltfx;

		// LED for drive 1
		if (GetOption(drive1))
		{
			bltfx.dwFillColor = g_Drive1.IsLightOn() ? dwOn : dwOff;
			pddsBack->Blt(&rLed, NULL, NULL, DDBLT_COLORFILL, &bltfx);
		}

		// LED for drive 2
		if (GetOption(drive2) == 1)
		{
			OffsetRect(&rLed, 8 * GetOption(sourcesize), 0);
			bltfx.dwFillColor = g_Drive2.IsLightOn() ? dwOn : dwOff;
			pddsBack->Blt(&rLed, NULL, NULL, DDBLT_COLORFILL, &bltfx);
		}
	}

	ProfileEnd();


	// Whether the next frame gets draw depends on the frame-skip setting
	fDrawFrame = GetOption(video) && (!nDrawnFrames || (GetOption(frameskip) ? !(nFrame % GetOption(frameskip)) :
					WaitForSingleObject(g_hEvent, 0) != WAIT_OBJECT_0));

	// Frame synchronisation required?
	ProfileStart(Idle);
	switch (GetOption(sync))
	{
		// Wait for timer to trigger the event
		case 1:
			if (g_hEvent)
				WaitForSingleObject(g_hEvent, INFINITE);
			break;

		// Use vsync - will give a smooth 60fps for a display mode with 60Hz vertical refresh
		// Shame most (or any!) video cards don't suppose a 50Hz screen mode
		case 2:
		{
			DWORD dwEnd = dwStartTime + (1000/EMULATED_FRAMES_PER_SECOND) - 2;

			// Vsync until we're within 2ms of the end of the frame - DOESN'T WORK PROPERLY YET!
//			DWORD dwNow;
//			while (((dwNow = timeGetTime()) < dwEnd))
				Video::WaitVBlank();
			break;
		}

		// No sync
		default:
			break;
	}
	ProfileEnd();

	ResetEvent(g_hEvent);
}

void Clear ()
{
	// Fill with black
	DDBLTFX bltfx;
	bltfx.dwSize = sizeof bltfx;
	bltfx.dwFillColor = 0;

	// Clear the back-buffer
	if (pddsBack)
		pddsBack->Blt(NULL, NULL, NULL, DDBLT_COLORFILL, &bltfx);

	// Clear the front-buffer
	if (pddsPrimary)
		pddsPrimary->Blt(NULL, NULL, NULL, DDBLT_COLORFILL, &bltfx);
}

void Redraw ()
{
	if (!pddsBack || !pddsFront || !pddsPrimary)
		return;

	pddsPrimary->Restore();
 	pddsBack->Restore();
	pddsFront->Restore();


/*
	if (GetOption(profile))
	{
		HDC hdc;
		pddsBack->GetDC(&hdc);
		HFONT hfontOld = reinterpret_cast<HFONT>(SelectObject(hdc, g_hfontStatus));

		int nOffset = GetOption(fullscreen) ? nXmargin : 0;

		// Draw a shadow in black text and then the main text in white, so it's legible on any background colour
		SetBkMode(hdc, TRANSPARENT);
		SetTextColor(hdc, RGB(0,0,0));
		TextOut(hdc, rectFrom.left+nOffset+1, rectFrom.top+1, g_szProfile, lstrlen(g_szProfile));
		SetTextColor(hdc, RGB(255,255,255));
		TextOut(hdc, rectFrom.left+nOffset, rectFrom.top, g_szProfile, lstrlen(g_szProfile));

		SelectObject(hdc, hfontOld);
		pddsBack->ReleaseDC(hdc);
	}
*/

	HRESULT hr;

	DDSURFACEDESC ddsdBack, ddsdFront, ddsdPrimary;
	ddsdBack.dwSize = ddsdFront.dwSize = ddsdPrimary.dwSize = sizeof ddsdBack;
	pddsBack->GetSurfaceDesc(&ddsdBack);
	pddsFront->GetSurfaceDesc(&ddsdFront);


	RECT rBack = { 0, 0, ddsdBack.dwWidth, ddsdBack.dwHeight };

	// If we're using the palette method we need to build the real image from the palette data prepared earlier
	if (GetOption(drawmethod) == 1)
	{
		ProfileStart(GFX);

		int nWidth = ddsdBack.dwWidth;
		int nHeight = ddsdBack.dwHeight;

		if (SUCCEEDED(hr = pddsBack->Lock(NULL, &ddsdBack, DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT, NULL)))
		{
			int nDepth = ddsdBack.ddpfPixelFormat.dwRGBBitCount;
			LONG lNextBack = (ddsdBack.lPitch - (nWidth * (nDepth >> 3))) / (nDepth >> 3);
			BYTE* pSAM = g_pDisplay->m_pbFrameStart;

			nWidth >>= 3;

			switch (nDepth)
			{
				case 8:
				{
					BYTE* pwBack = reinterpret_cast<BYTE*>(ddsdBack.lpSurface);

					for (int y = 0 ; y < nHeight ; y++)
					{
						for (int x = 0 ; x < nWidth ; x++)
						{
							// Draw 8 pixels at a time
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
							*pwBack++ = (BYTE)aulPalette[*pSAM++];
						}
						pwBack += lNextBack;
					}
				}
				break;

				case 16:
				{
					WORD* pwBack = reinterpret_cast<WORD*>(ddsdBack.lpSurface);

					for (int y = 0 ; y < nHeight ; y++)
					{
						for (int x = 0 ; x < nWidth ; x++)
						{
							// Draw 8 pixels at a time
							*pwBack++ = (WORD)aulPalette[*pSAM++];
							*pwBack++ = (WORD)aulPalette[*pSAM++];
							*pwBack++ = (WORD)aulPalette[*pSAM++];
							*pwBack++ = (WORD)aulPalette[*pSAM++];
							*pwBack++ = (WORD)aulPalette[*pSAM++];
							*pwBack++ = (WORD)aulPalette[*pSAM++];
							*pwBack++ = (WORD)aulPalette[*pSAM++];
							*pwBack++ = (WORD)aulPalette[*pSAM++];
						}
						pwBack += lNextBack;
					}
				}
				break;

				case 32:
				{
					DWORD* pwBack = reinterpret_cast<DWORD*>(ddsdBack.lpSurface);

					for (int y = 0 ; y < nHeight ; y++)
					{
						for (int x = 0 ; x < nWidth ; x++)
						{
							// Draw 8 pixels at a time
							*pwBack++ = aulPalette[*pSAM++];
							*pwBack++ = aulPalette[*pSAM++];
							*pwBack++ = aulPalette[*pSAM++];
							*pwBack++ = aulPalette[*pSAM++];
							*pwBack++ = aulPalette[*pSAM++];
							*pwBack++ = aulPalette[*pSAM++];
							*pwBack++ = aulPalette[*pSAM++];
							*pwBack++ = aulPalette[*pSAM++];
						}
						pwBack += lNextBack;
					}
				}
				break;
			}

			pddsBack->Unlock(ddsdBack.lpSurface);
		}

		ProfileEnd();
	}


	ProfileStart(Blit);

	// If the front buffer isn't the primary buffer we need to do a non-stretched copy into it
	// We do this manually as it should be faster than BltFast, and because a YUV blit may not work
	if (pddsBack != pddsFront && pddsFront != pddsPrimary)
	{
		if (SUCCEEDED(hr = pddsBack->Lock(NULL, &ddsdBack, DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT, NULL)))
		{
			BYTE* pbBack = reinterpret_cast<BYTE*>(ddsdBack.lpSurface);
			LONG lPitchBack = ddsdBack.lPitch;

			if (SUCCEEDED(hr = pddsFront->Lock(NULL, &ddsdFront, DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT, NULL)))
			{
				BYTE* pbFront = reinterpret_cast<BYTE*>(ddsdFront.lpSurface);
				LONG lPitchFront = ddsdFront.lPitch;

				int nWidthBytes = (rBack.right - rBack.left) * (ddsdBack.ddpfPixelFormat.dwRGBBitCount >> 3);
				int nHeight = rBack.bottom - rBack.top;

				for (int i = 0 ; i < nHeight ; i++)
				{
					memcpy(pbFront, pbBack, nWidthBytes);
					pbBack += lPitchBack;
					pbFront += lPitchFront;
				}

				pddsFront->Unlock(ddsdFront.lpSurface);
			}

			pddsBack->Unlock(ddsdBack.lpSurface);
		}
	}


	// If the front buffer is NOT an overlay we need to blit it to the primary surface to get it displayed
	RECT rPrimary;
	pddsFront->GetSurfaceDesc(&ddsdFront);
	if (!(ddsdFront.ddsCaps.dwCaps & DDSCAPS_OVERLAY))
	{
		if (GetOption(fullscreen))
		{
			// Get details on the primary surface
			pddsPrimary->GetSurfaceDesc(&ddsdPrimary);

			DWORD dwWidth = GetOption(ratio5_4) ? ddsdBack.dwWidth * 5 / 4 : ddsdBack.dwWidth;
			DWORD dwHeight = ddsdBack.dwHeight * GetOption(sourcesize);

			// If the target surface is too wide to display, adjust the portion of the source that we're viewing
			if (dwWidth > ddsdPrimary.dwWidth)
			{
				int nDiff = (ddsdBack.dwWidth - MulDiv(ddsdBack.dwWidth, ddsdPrimary.dwWidth, dwWidth)) >> 1;
				rBack.left += nDiff;
				rBack.right -= nDiff;
				dwWidth = ddsdPrimary.dwWidth;
			}

			// If the target surface is too high to display, adjust the portion of the source that we're viewing
			if (dwHeight > ddsdPrimary.dwHeight)
			{
				int nDiff = (ddsdBack.dwHeight - MulDiv(ddsdBack.dwHeight, ddsdPrimary.dwHeight, dwHeight)) >> 1;
				rBack.top += nDiff;
				rBack.bottom -= nDiff;
				dwHeight = ddsdPrimary.dwHeight;
			}

			// Work out offsets required to centralise the target image, and work out the final rectangle
			int nOffsetX = (ddsdPrimary.dwWidth - dwWidth) >> 1, nOffsetY = (ddsdPrimary.dwHeight - dwHeight) >> 1;
			SetRect(&rPrimary, nOffsetX, nOffsetY, nOffsetX + dwWidth, nOffsetY + dwHeight);
		}

		// Windowed output only displays the visible part
		else
		{
			// The target image is to fill the client area of the window (which is already the correct size)
			GetClientRect(g_hwnd, &rPrimary);

			// Offset the rectange to position where the top left of the client area is on the screen
			POINT pt = {0,0};
			ClientToScreen(g_hwnd, &pt);
			OffsetRect(&rPrimary, pt.x, pt.y);
		}

		if (FAILED(hr = pddsPrimary->Blt(&rPrimary, pddsBack, NULL, DDBLT_WAIT, 0)))
			TRACE("!!! Blt failed with %#08lx\n", hr);
	}
	else
	{
		// Fill the primary surface area with the colour key
		DDBLTFX bltfx;
		bltfx.dwSize = sizeof bltfx;
		bltfx.dwFillColor = g_dwColourKey;
		pddsPrimary->Blt(NULL, NULL, NULL, DDBLT_COLORFILL, &bltfx);
	}

	ProfileEnd();
}

}	// namespace Frame
