//	Win32 specific sound implementation using DirectSound to play the data produced by SAASOUND.DLL
//
//	ToDo:	Explicitly load and unload SAASOUND.DLL so SimCoupe can run without it, as
//			SAASound is not released under the same licence and needs to be optional.

#include "stdafx.h"

#include <dsound.h>

#include "IO.h"
#include "CPU.h"
#include "Sound.h"
#include "Options.h"
#include "Util.h"
#include "SAASound.h"
#include "Parallel.h"
#include "Profile.h"

extern HWND g_hwnd;
extern int g_nLine, g_nLineCycle;

CStreamingSound *pDAC1, *pDAC2;

// Direct sound, primary buffer and secondary buffer interface pointers
IDirectSound* CStreamingSound::s_pds = NULL;
IDirectSoundBuffer* CStreamingSound::s_pdsbPrimary;
int CStreamingSound::s_nUsage = 0;


class CSAASoundDirectX : public CSoundDriver, public CStreamingSound
{
	public:
		~CSAASoundDirectX ()
		{
			if (m_pSAASound)
				DestroyCSAASound(m_pSAASound);
		}

	public:
		bool Init ()
		{
			bool fRet = CStreamingSound::Init() && (m_pSAASound = CreateCSAASound());

			if (m_pSAASound)
			{
				m_pSAASound->SetSoundParameters(
					(GetOption(filter) ? SAAP_FILTER : SAAP_NOFILTER) |
					(GetOption(rate) < 20000 ? SAAP_11025 : GetOption(rate) < 40000 ? SAAP_22050 : SAAP_44100) |
					(GetOption(bits) < 12 ? SAAP_8BIT : SAAP_16BIT) |
					(GetOption(stereo) ? SAAP_STEREO : SAAP_MONO));
			}

			return fRet;
		}

	public:
		void Generate (BYTE* pb_, UINT uSamples_)
		{
			m_pSAASound->GenerateMany(pb_, uSamples_);
		}

		void GenerateExtra (BYTE* pb_, UINT uSamples_)
		{
			// DAC optimisation will generate extra samples using the default method of repeating
			// This should roughly continue the same frequency as before
			if (GetOption(dacoptimise))
				CStreamingSound::GenerateExtra(pb_, uSamples_);
//ADD FROM MOST RECENT DATA SECTION INSTEAD?
//ADD TO UI


			// Else continue the SAA sample data to give a smooth addition to what we've already got
			else
				m_pSAASound->GenerateMany(pb_, uSamples_);
		}

		void Out (WORD wPort_, BYTE bVal_)
		{
			Update();

			if ((wPort_ & SOUND_MASK) == SOUND_ADDR)
				m_pSAASound->WriteAddress(bVal_);
			else
				m_pSAASound->WriteData(bVal_);
		}

		void Click (BYTE bVal_)
		{
			Update();
			m_pSAASound->ClickClick(bVal_);
		}

		void Update (bool fFrameEnd_=false) { CStreamingSound::Update(fFrameEnd_); }
		bool Play ()	{ return CStreamingSound::Play(); }
		bool Stop ()	{ CStreamingSound::Silence(); return CStreamingSound::Stop(); }
		void Silence ()	{ CStreamingSound::Silence(); }

	protected:
		LPCSAASOUND m_pSAASound;
};


namespace Sound
{

CSoundDriver* pDriver;		// Pointer to the current driver object for dealing with the sound chip
BYTE abRegs[N_SAA_REGS];	// We need to keep track of the SAA 1099 register values to update them after a driver change

////////////////////////////////////////////////////////////////////////////////

bool Init ()
{
	// Clear out any existing config before starting again
	Exit();

	// Create the appropriate sound driver object
	switch (GetOption(sound))
	{
		case 1:		pDriver = new CSAASoundDirectX;	break;

		default:
			TRACE("Sound disabled, nothing to initialise\n");
			pDriver = NULL;
			break;
	}

	if (pDriver)
	{
		if (!pDriver->Init())
		{
			Message(msgWarning, "Sound initialisation failed, disabling...");
			Exit();
			SetOption(sound,false);
		}
		else
		{
			// Re-initialise the sound-chip registers with the last known values
			for (int i = N_SAA_REGS-1 ; i >= 0 ; i--)
			{
				pDriver->Out(SOUND_ADDR, i);
				pDriver->Out(SOUND_DATA, abRegs[i]);
			}
		}
	}


	switch (GetOption(parallel1))
	{
		case 2:		pDAC1 = new CMonoDAC;	break;
		case 3:		pDAC1 = new CStereoDAC;	break;
	}

	if (pDAC1 && !pDAC1->Init())
	{
		delete pDAC1;
		pDAC1 = NULL;

		Message(msgWarning, "DAC initialisation failed, disabling...");
		SetOption(parallel1, 0);
	}


	switch (GetOption(parallel2))
	{
		case 2:		pDAC2 = new CMonoDAC;	break;
		case 3:		pDAC2 = new CStereoDAC;	break;
	}

	if (pDAC2 && !pDAC2->Init())
	{
		delete pDAC2;
		pDAC2 = NULL;

		Message(msgWarning, "DAC2 initialisation failed, disabling...");
		SetOption(parallel2, 0);
	}


	// Sound initialisation failure isn't fatal, so always return success
	return true;
}

void Exit (bool fReinit_/*=false*/)
{
	if (pDriver) { delete pDriver; pDriver = NULL; }
	if (pDAC1)	{ delete pDAC1; pDAC1 = NULL; }
	if (pDAC2)	{ delete pDAC2; pDAC2 = NULL; }
}

void Out (WORD wPort_, BYTE bVal_)
{
	static BYTE bReg;

	if (pDriver)
	{
		if ((wPort_ & SOUND_MASK) == SOUND_ADDR)
			bReg = bVal_ & (N_SAA_REGS-1);
		else
			abRegs[bReg] = bVal_;

		pDriver->Out(wPort_, bVal_);
	}
}

void Click (int bVal_)
{
	ProfileStart(Sound);

	if (pDriver)
		pDriver->Click(bVal_);

	ProfileEnd();
}


void FrameUpdate ()
{
	ProfileStart(Sound);

	if (pDriver) pDriver->Update(true);
	if (pDAC1) pDAC1->Update(true);
	if (pDAC2) pDAC2->Update(true);

	ProfileEnd();
}


void Silence ()
{
	if (pDriver) pDriver->Silence();
	if (pDAC1) pDAC1->Silence();
	if (pDAC2) pDAC2->Silence();
}

void Stop ()
{
	if (pDriver) { pDriver->Stop(); pDriver->Silence(); }
	if (pDAC1) { pDAC1->Stop(); pDAC1->Silence(); }
	if (pDAC2) { pDAC2->Stop(); pDAC2->Silence(); }
}

void Play ()
{
	if (pDriver) pDriver->Play();
	if (pDAC1) pDAC1->Play();
	if (pDAC2) pDAC2->Play();
}

};	// namespace Sound




bool CStreamingSound::InitDirectSound ()
{
	HRESULT hr;
	bool fRet = false;

	// Initialise DirectX
	if (FAILED(hr = DirectSoundCreate(NULL, &s_pds, NULL)))
		TRACE("!!! DirectSoundCreate failed (%#08lx)\n", hr);

	// We want priority control over the sound format while we're active
	else if (FAILED(hr = s_pds->SetCooperativeLevel(g_hwnd, DSSCL_PRIORITY)))
		TRACE("!!! SetCooperativeLevel() failed (%#08lx)\n", hr);
	else
	{
		DSBUFFERDESC dsbd;
		ZeroMemory(&dsbd, sizeof dsbd);
		dsbd.dwSize = sizeof DSBUFFERDESC;
		dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;

		// Create the primary buffer
		if (FAILED(hr = s_pds->CreateSoundBuffer(&dsbd, &s_pdsbPrimary, NULL)))
			TRACE("!!! CreateSoundBuffer() failed with primary buffer (%#08lx)\n", hr);

		// Eliminate the mixing slowdown by setting the primary buffer to our required format
		else
		{
			// Set up the sound format according to the sound options
			WAVEFORMATEX wf;
			wf.cbSize = sizeof WAVEFORMATEX;
			wf.wFormatTag = WAVE_FORMAT_PCM;
			wf.nChannels = GetOption(stereo) ? 2 : 1;
			wf.wBitsPerSample = GetOption(bits) > 8 ? 16 : 8;
			wf.nBlockAlign = wf.nChannels * wf.wBitsPerSample / 8;
			wf.nSamplesPerSec = GetOption(rate) < 20000 ? 11025 : GetOption(rate) < 40000 ? 22050 : 44100;
			wf.nAvgBytesPerSec = wf.nSamplesPerSec * wf.nBlockAlign;

			// Set the primary buffer format (closest match will be used if hardware doesn't support what we request)
			if (FAILED(hr = s_pdsbPrimary->SetFormat(&wf)))
				TRACE("!!! SetFormat() failed on primary buffer (%#08lx)\n", hr);

			// Play the primary buffer all the time to eliminate the mixer delays in stopping and starting
			else if (FAILED(hr = s_pdsbPrimary->Play(0, 0, DSBPLAY_LOOPING)))
				TRACE("!!! Play() failed on primary buffer (%#08lx)\n", hr);

			// Success :-)
			else
				fRet = true;
		}
	}

	return fRet;
}


void CStreamingSound::ExitDirectSound ()
{
	if (s_pdsbPrimary)
	{
		s_pdsbPrimary->Stop();
		s_pdsbPrimary->Release();
		s_pdsbPrimary = NULL;
	}

	if (s_pds)
	{
		s_pds->Release();
		s_pds = NULL;
	}
}



CStreamingSound::CStreamingSound (int nFreq_/*=0*/, int nBits_/*=0*/, int nChannels_/*=0*/)
{
	m_nFreq = nFreq_;
	m_nBits = nBits_;
	m_nChannels = nChannels_;
}

CStreamingSound::~CStreamingSound ()
{
	if (m_pdsb) { m_pdsb->Release(); m_pdsb = NULL; }
	if (m_pbFrameSample) { delete m_pbFrameSample; m_pbFrameSample = NULL; }

	// If this is the last object we have to close down DirectSound
	if (!--s_nUsage)
		ExitDirectSound();
}


bool CStreamingSound::Init ()
{
	bool fRet = true;

	// If this is the first object we have to set up DirectSound
	if (!s_nUsage++)
		fRet = InitDirectSound();

	if (fRet)
	{
		HRESULT hr;

		WAVEFORMATEX wf;
		s_pdsbPrimary->GetFormat(&wf,  sizeof wf, NULL);
		wf.cbSize = sizeof wf;

		if (!m_nFreq)
			m_nFreq = wf.nSamplesPerSec;

		if (!m_nBits)
			m_nBits = wf.wBitsPerSample;

		if (!m_nChannels)
			m_nChannels = wf.nChannels;

		wf.nChannels = m_nChannels;
		wf.wBitsPerSample = m_nBits;
		wf.nSamplesPerSec = m_nFreq;

		wf.nBlockAlign = m_nChannels * m_nBits / 8;
		wf.nAvgBytesPerSec = (m_nFreq) * wf.nBlockAlign;


		m_uSamplesPerFrame = m_nFreq / EMULATED_FRAMES_PER_SECOND;
		m_uSampleSize = m_nChannels * m_nBits / 8;
		m_uSampleBufferSize = (m_uSamplesPerFrame * m_uSampleSize * GetOption(latency));
//		m_uSampleBufferSize = (m_uSampleBufferSize + (16-1)) & ~(16-1);

		m_pbFrameSample = new BYTE[m_uSampleBufferSize];
		m_uSamplesThisFrame = 0;
		m_dwWriteOffset = 0;


		DSBUFFERDESC dsbd;
		ZeroMemory(&dsbd, sizeof dsbd);
		dsbd.dwSize = sizeof DSBUFFERDESC;
		dsbd.dwFlags = DSBCAPS_GETCURRENTPOSITION2;// | DSBCAPS_GLOBALFOCUS;
		dsbd.dwBufferBytes = m_uSampleBufferSize;
		dsbd.lpwfxFormat = &wf;

		m_pdsb = NULL;
		if (FAILED(hr = s_pds->CreateSoundBuffer(&dsbd, &m_pdsb, NULL)))
			TRACE("!!! CreateSoundBuffer failed (%#08lx)\n", hr);
		else
		{
			// Make sure the buffer contents is silent before playing
			Silence();

			// Play for as long as we're running
			if (FAILED(hr = m_pdsb->Play(0, 0, DSBPLAY_LOOPING)))
				TRACE("!!! Play failed on DirectSound buffer (%#08lx)\n", hr);
		}
	}

	return fRet;
}


void CStreamingSound::Silence ()
{
	PVOID pvWrite1, pvWrite2;
	DWORD dwLength1, dwLength2;

	// Lock the entire buffer and fill will the values for silence
	if (SUCCEEDED(m_pdsb->Lock(0, 0, &pvWrite1, &dwLength1, &pvWrite2, &dwLength2, DSBLOCK_ENTIREBUFFER)))
	{
		FillMemory(pvWrite1, dwLength1, (m_nBits == 16) ? 0x00 : 0x80);
		m_pdsb->Unlock(pvWrite1, dwLength1, pvWrite2, dwLength2);
	}

	// The new write offset will be the current write cursor position
	DWORD dwPlayCursor, dwWriteCursor;
	m_pdsb->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor);
	m_dwWriteOffset = dwWriteCursor;
}


DWORD CStreamingSound::GetSpaceAvailable ()
{
	HRESULT hr;
	DWORD dwPlayCursor, dwWriteCursor;

	// Get the current play cursor position
	if (FAILED(hr = m_pdsb->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor)))
		return 0;

	// The amount of space free depends on where the last write postion is relative to the play cursor
	UINT uSpace = (m_dwWriteOffset <= dwPlayCursor) ? dwPlayCursor - m_dwWriteOffset :
					m_uSampleBufferSize - (m_dwWriteOffset - dwPlayCursor);

	// Return the number of samples
	return uSpace / m_uSampleSize;
}


void CStreamingSound::Update (bool fFrameEnd_)
{
    ProfileStart(Sound);

    // Calculate the number of samples from the start of the frame to the current position
    UINT uCurrentCycles = (g_nLine * TSTATES_PER_LINE) + g_nLineCycle;
    UINT uSamplesSoFar = MulDiv(m_uSamplesPerFrame, uCurrentCycles, TSTATES_PER_FRAME);

    // Generate and append the the additional sample(s) to our temporary buffer
    if (uSamplesSoFar > m_uSamplesThisFrame)
    {
        Generate(m_pbFrameSample + (m_uSamplesThisFrame * m_uSampleSize), uSamplesSoFar - m_uSamplesThisFrame);
        m_uSamplesThisFrame = uSamplesSoFar;
    }

    if (fFrameEnd_)
    {
		DWORD dwSpaceAvailable = GetSpaceAvailable();

		// Is there enough space for all this frame's data?
		if (dwSpaceAvailable >= m_uSamplesThisFrame)
		{
			// Add on the current frame's sample data
			AddData(m_pbFrameSample, m_uSamplesThisFrame);
			dwSpaceAvailable -= m_uSamplesThisFrame;

			// Have we fallen below the hover range?
			if (dwSpaceAvailable > m_uSamplesPerFrame)
			{
//				OutputDebugString("Below hover range\n");

				// Calculate the remaining space below the hover point
				dwSpaceAvailable -= (m_uSamplesPerFrame >> 1);

				// Add as many additional full frames as are needed to get close to the hover point (without exceeding it)
				while (dwSpaceAvailable >= m_uSamplesPerFrame)
				{
//					OutputDebugString(" Added generated frame\n");
					GenerateExtra(m_pbFrameSample, m_uSamplesPerFrame);
					AddData(m_pbFrameSample, m_uSamplesPerFrame);
					dwSpaceAvailable -= m_uSamplesPerFrame;
				}

				// Top up the buffer to the hover point
				if (dwSpaceAvailable)
				{
//					OutputDebugString(" Added part generated frame\n");
					GenerateExtra(m_pbFrameSample, dwSpaceAvailable);
					AddData(m_pbFrameSample, dwSpaceAvailable);
				}
			}
		}

		// Else there's not enough space for the full frame of data, but we'll add what we can to leave us at the hover point
		else if (dwSpaceAvailable >= (m_uSamplesPerFrame >> 1))
		{
//			OutputDebugString("Above hover range\n");
//			OutputDebugString(" Adding part frame data\n");
			AddData(m_pbFrameSample, dwSpaceAvailable - (m_uSamplesPerFrame >> 1));
		}


        // Reset the sample counter for the next frame
		m_uSamplesThisFrame = 0;

    }

    ProfileEnd();
}


void CStreamingSound::AddData (BYTE* pbSampleData_, UINT uSamples_)
{
	LPVOID pvWrite1, pvWrite2;
	DWORD dwLength1, dwLength2;

	// Lock the required buffer range so we can add the new data
	HRESULT hr = m_pdsb->Lock(m_dwWriteOffset, uSamples_ * m_uSampleSize, &pvWrite1, &dwLength1, &pvWrite2, &dwLength2, 0);
	if (FAILED(hr))
		TRACE("!!! Failed to lock sound buffer! (%#08lx)\n", hr);
	else
	{
		// Write the first part (maybe all)
		memcpy(pvWrite1, pbSampleData_, dwLength1);

		// Write the second block if necessary
		if (dwLength2)
			memcpy(pvWrite2, pbSampleData_+dwLength1, dwLength2);

		// Unlock the buffer
		hr = m_pdsb->Unlock(pvWrite1, dwLength1, pvWrite2, dwLength2);
		if (FAILED(hr))
			TRACE("!!! Failed to unlock sound buffer! (%#08lx)\n", hr);

		// Work out the offset for the next write in the circular buffer
		m_dwWriteOffset = (m_dwWriteOffset + dwLength1 + dwLength2) % m_uSampleBufferSize;
	}
}


void CStreamingSound::GenerateExtra (BYTE* pb_, UINT uSamples_)
{
	// Re-use the specified amount from the previous sample, taking care of buffer overlaps
	if (pb_ != m_pbFrameSample)
		memmove(pb_, m_pbFrameSample, uSamples_*m_uSampleSize);
}
